import attrs
from .base import Field


@attrs.define
class Str(Field):

    value: str = attrs.field(converter=str, default=None)


@attrs.define
class Bool(Field):

    value: bool = attrs.field(converter=bool, default=None)


@attrs.define
class Int(Field):

    value: int = attrs.field(converter=int, default=None)


@attrs.define
class Float(Field):

    value: float = attrs.field(converter=float, default=None)



from pathlib import Path
import click
import arcana
import arcana.core as core
from arcana.core.utils.packaging import submodules


# Define the base CLI entrypoint
@click.group()
@click.version_option(version=core.__version__)
def cli():
    """Base command line group, installed as "arcana"."""


@cli.group()
def ext():
    """Command-line group for extension hooks"""


CLI_EXT_PKG = Path(core.__file__).parent / "cli"

# Ensure that all sub-packages under CLI are loaded so they are added to the
# base command
submodules(arcana, subpkg="cli")
# for module in pkgutil.iter_modules([str(CLI_EXT_PKG)], prefix="arcana.core.cli."):
#     import_module(module.name)



import typing as ty
import json
from pydra import mark
from pydra.engine.core import File
from pydra.engine.specs import BaseSpec, SpecInfo
from pydra.engine.task import FunctionTask
from arcana.core.data.type.base import DataType, FileGroup
from arcana.core.data.row import DataRow


def identity(**fields):
    return fields


def identity_task(task_name, fields):
    task = FunctionTask(
        identity,
        input_spec=SpecInfo(
            name=f"{task_name}Inputs",
            bases=(BaseSpec,),
            fields=[(s, DataType) for s in fields],
        ),
        output_spec=SpecInfo(
            name=f"{task_name}Outputs", bases=(BaseSpec,), fields=[("row", DataRow)]
        ),
    )
    return task


@mark.task
@mark.annotate({"in_file": FileGroup, "return": {"out_file": FileGroup}})
def identity_converter(in_file):
    return in_file


@mark.task
def extract_from_json(in_file: File, field_name: str) -> ty.Any:
    with open(in_file) as f:
        dct = json.load(f)
    return dct[field_name]  # FIXME: Should use JSONpath syntax



import json
import tempfile
import attrs
from pathlib import Path
from .base import ArcanaImage


@attrs.define
class Metapackage(ArcanaImage):

    manifest: dict

    def construct_dockerfile(
        self,
        build_dir: Path = None,
        use_local_packages=False,
    ):

        if build_dir is None:
            build_dir = Path(tempfile.mkdtemp())

        dockerfile = self.init_dockerfile()

        self.install_python(
            dockerfile, build_dir, use_local_packages=use_local_packages
        )

        with open(build_dir / "manifest.json", "w") as f:
            json.dump(self.manifest, f)

        dockerfile.copy(["./manifest.json"], "/manifest.json")

        dockerfile.entrypoint(
            [
                "conda",
                "run",
                "--no-capture-output",
                "-n",
                "arcana",
                "arcana",
                "ext",
                "xnat",
                "pull-images",
                "/manifest.json",
            ]
        )

        return dockerfile



from operator import mul
from functools import reduce
from arcana.core.data.type.base import FileGroup


def test_column_api_access(dataset):

    bp = dataset.__annotations__["blueprint"]

    for col_name, formats in bp.expected_formats.items():
        data_format, files = formats[0]

        dataset.add_source(col_name, data_format)

        col = dataset[col_name]

        # Check length of column
        assert len(col) == reduce(mul, bp.dim_lengths)

        # Access file-group via leaf IDs
        for id_ in col.ids:
            item = col[id_]
            assert isinstance(item, data_format)
            if issubclass(data_format, FileGroup):
                assert sorted(p.name for p in item.fs_paths) == sorted(files)



from __future__ import annotations
import os
from pathlib import Path
import typing as ty
from itertools import chain
from copy import copy
import shutil
import attrs
from arcana.core.exceptions import (
    ArcanaUsageError,
    ArcanaFileFormatError,
)
from .base import FileGroup, absolute_path


@attrs.define
class BaseDirectory(FileGroup):

    is_dir = True
    content_types = ()  # By default, don't check contents for any types

    def set_fs_paths(self, fs_paths: ty.List[Path]):
        self._check_paths_exist(fs_paths)
        matches = [p for p in fs_paths if Path(p).is_dir() and self.contents_match(p)]
        types_str = ", ".join(t.__name__ for t in self.content_types)
        if not matches:
            raise ArcanaFileFormatError(
                f"No matching directories with contents matching {types_str} amongst "
                f"{fs_paths}"
            )
        elif len(matches) > 1:
            matches_str = ", ".join(str(m) for m in matches)
            raise ArcanaFileFormatError(
                f"Multiple directories with contents matching {types_str}: "
                f"{matches_str}"
            )
        self.exists = True
        self.fs_path = absolute_path(matches[0])

    @classmethod
    def contents_match(cls, path: Path):
        from arcana.core.data.row import UnresolvedFileGroup

        path = Path(path)  # Ensure a Path object not a string
        contents = UnresolvedFileGroup.from_paths(path, path.iterdir())
        for content_type in cls.content_types:
            resolved = False
            for unresolved in contents:
                try:
                    content_type.resolve(unresolved)
                except ArcanaFileFormatError:
                    pass
                else:
                    resolved = True
                    break
            if not resolved:
                return False
        return True

    def all_file_paths(self):
        "Iterates through all files in the group and returns their file paths"
        if self.fs_path is None:
            raise ArcanaUsageError(
                f"Attempting to access file paths of {self} before they are set"
            )
        return chain(
            *(
                (Path(root) / f for f in files)
                for root, _, files in os.walk(self.fs_path)
            )
        )

    def copy_to(self, fs_path: str, symlink: bool = False):
        """Copies the file-group to the new path, with auxiliary files saved
        alongside the primary-file path.

        Parameters
        ----------
        fs_path : str
            Path to save the file-group to excluding file extensions
        symlink : bool
            Use symbolic links instead of copying files to new location
        """
        if symlink:
            copy_dir = os.symlink
        else:
            copy_dir = shutil.copytree
        copy_dir(self.fs_path, fs_path)
        cpy = copy(self)
        cpy.set_fs_paths([fs_path])
        return cpy



from __future__ import annotations
from typing import Sequence
import json
import importlib_metadata
import pkgutil
from importlib import import_module
from inspect import isclass
import pkg_resources
from pathlib import Path
from collections.abc import Iterable
from arcana.core.exceptions import ArcanaUsageError
from arcana.core import __version__


def submodules(package, subpkg=None):
    """Iterates all modules within the given package

    Parameters
    ----------
    package : module
        the package to iterate over
    subpkg : str, optional
        the sub-package (of the sub-packages) to return instead of the first level down.
        e.g. package=arcana, subpkg=data -> arcana.dirtree.data, arcana.xnat.data, etc...

    Yields
    ------
    module
        all modules within the package
    """
    for mod_info in pkgutil.iter_modules(
        package.__path__, prefix=package.__package__ + "."
    ):
        if subpkg is not None:
            try:
                yield import_module(mod_info.name + "." + subpkg)
            except ImportError:
                continue
        else:
            yield import_module(mod_info.name)


def list_subclasses(package, base_class, subpkg=None):
    """List all available subclasses of a base class in modules within the given
    package

    Parameters
    ----------
    package : module
        the package to list the subclasses within
    base_class : type
        the base class
    subpkg : str, optional
        the sub-package (of the sub-packages) to return instead of the first level down.
        e.g. package=arcana, subpkg=data -> arcana.dirtree.data, arcana.xnat.data, etc...

    Returns
    -------
    list
        all subclasses of the base-class found with the package
    """
    subclasses = []
    for module in submodules(package, subpkg=subpkg):
        for obj_name in dir(module):
            obj = getattr(module, obj_name)
            if isclass(obj) and issubclass(obj, base_class) and obj is not base_class:
                subclasses.append(obj)
    return subclasses


def package_from_module(module: Sequence[str]):
    """Resolves the installed package (e.g. from PyPI) that provides the given
    module.

    Parameters
    ----------
    module: str or module or Sequence[str or module]
        a module or its import path string to retrieve the package for. Can be
        provided as a list of modules/strings, in which case a list of packages
        are returned

    Returns
    -------
    PackageInfo or list[PackageInfo]
        the package info object corresponding to the module. If `module`
        parameter is a list of modules/strings then a set of packages are
        returned
    """
    module_paths = set()
    if isinstance(module, Iterable) and not isinstance(module, str):
        modules = module
        as_tuple = True
    else:
        modules = [module]
        as_tuple = False
    for module in modules:
        try:
            module_path = module.__name__
        except AttributeError:
            module_path = module
        module_paths.add(importlib_metadata.PackagePath(module_path.replace(".", "/")))
    packages = set()
    for pkg in pkg_resources.working_set:
        if editable_dir := get_editable_dir(pkg):

            def is_in_pkg(module_path):
                pth = editable_dir.joinpath(module_path)
                return pth.with_suffix(".py").exists() or (pth / "__init__.py").exists()

        else:
            installed_paths = installed_module_paths(pkg)

            def is_in_pkg(module_path):
                return module_path in installed_paths

        if in_pkg := set(m for m in module_paths if is_in_pkg(m)):
            packages.add(pkg)
            module_paths -= in_pkg
            if not module_paths:  # If there are no more modules to find pkgs for break
                break
    if module_paths:
        paths_str = "', '".join(str(p) for p in module_paths)
        raise ArcanaUsageError(f"Did not find package for {paths_str}")
    return tuple(packages) if as_tuple else next(iter(packages))


def get_editable_dir(pkg: pkg_resources.DistInfoDistribution):
    """Returns the path to the editable dir to a package if it exists

    Parameters
    ----------
    pkg : pkg_resources.DistInfoDistribution
        the package to get the editable directory for

    Returns
    ------
    Path or None
        the path to the editable file or None if the package isn't installed in editable mode
    """
    direct_url_path = Path(pkg.egg_info) / "direct_url.json"
    if not direct_url_path.exists():
        return None
    with open(direct_url_path) as f:
        url_spec = json.load(f)
    url = url_spec["url"]
    if not url_spec["dir_info"].get("editable"):
        return None
    assert url.startswith("file://")
    return Path(url[len("file://") :])


def installed_module_paths(pkg: pkg_resources.DistInfoDistribution):
    """Returns the list of modules that are part of an installed package

    Parameters
    ----------
    pkg
        the package to list the installed modules
    """
    try:
        paths = importlib_metadata.files(pkg.key)
    except importlib_metadata.PackageNotFoundError:
        paths = []
    paths = set(
        p.parent if p.name == "__init__.py" else p.with_suffix("")
        for p in paths
        if p.suffix == ".py"
    )
    return paths


def pkg_versions(modules):
    versions = {p.key: p.version for p in package_from_module(modules)}
    versions["arcana"] = __version__
    return versions



from __future__ import annotations
from pathlib import Path
import click
from click_option_group import optgroup
from arcana.core.deploy.image import App


def data_columns(func):

    return _apply_options(
        func,
        [
            optgroup.group(
                "Data columns",
                help=(
                    "Configure data columns to be added to the dataset before "
                    "the application and execution of the pipeline"
                ),
            ),
            optgroup.option(
                "--input",
                "-i",
                "input_values",
                nargs=2,
                default=(),
                metavar="<col-name> <match-criteria>",
                multiple=True,
                type=str,
                help=("The match criteria to pass to the column"),
            ),
            optgroup.option(
                "--output",
                "-o",
                "output_values",
                nargs=2,
                default=(),
                metavar="<col-name> <output-path>",
                multiple=True,
                type=str,
                help=("The path in which to store the output of the pipeline"),
            ),
        ],
    )


def parameterisation(func):
    return _apply_options(
        func,
        [
            optgroup.group(
                "Parameterisation",
                help=("Parameterisation of the pipeline and related options "),
            ),
            optgroup.option(
                "--parameter",
                "-p",
                "parameter_values",
                nargs=2,
                default=(),
                metavar="<name> <value>",
                multiple=True,
                type=str,
                help=("sets a parameter of the workflow"),
            ),
            optgroup.option(
                "--dataset-name",
                type=str,
                default=None,
                help=(
                    "The name of the dataset. Will be created if not present. Dataset names "
                    "are used to separate different analyses performed on the same data "
                    "into different namespaces"
                ),
            ),
            optgroup.option(
                "--overwrite/--no-overwrite",
                type=bool,
                help=(
                    "Whether to overwrite a saved pipeline with the same name, and its "
                    "parameterisation, if present"
                ),
            ),
        ],
    )


def execution(func):
    return _apply_options(
        func,
        [
            optgroup.group(
                "Execution", help=("Control over how the pipelines are executed")
            ),
            optgroup.option(
                "--work",
                "-w",
                "work_dir",
                default=None,
                help=(
                    "The location of the directory where the working files "
                    "created during the pipeline execution will be stored"
                ),
            ),
            optgroup.option(
                "--export-work",
                default=None,
                type=click.Path(exists=False, path_type=Path),
                help=(
                    "Export the work directory to another location after the task/workflow "
                    "exits (used for post-hoc investigation of completed workflows in "
                    "situations where the scratch space is inaccessible after the "
                    "workflow exits"
                ),
            ),
            optgroup.option(
                "--plugin",
                default="cf",
                help=("The Pydra plugin with which to process the task/workflow"),
            ),
            optgroup.option(
                "--loglevel",
                type=str,
                default="info",
                help=("The level of detail logging information is presented"),
            ),
            optgroup.option(
                "--ids",
                default=None,
                type=str,
                help=(
                    "List of IDs to restrict the pipeline execution to (i.e. don't execute "
                    "over the whole dataset)"
                ),
            ),
            optgroup.option(
                "--single-row",
                type=str,
                default=None,
                help=(
                    "Restrict the dataset created to a single row to avoid reduce start up "
                    "times and avoid ssues with unrelated rows that aren't being processed. "
                    "Comma-separated list of IDs for each layer of the hierarchy, e.g. "
                    "--single-row mysubject,mysession"
                ),
            ),
        ],
    )


def dataset_config(func):
    return _apply_options(
        func,
        [
            optgroup.group(
                "Dataset configuration",
                help=(
                    "Provide additional information to control how the dataset is created"
                ),
            ),
            optgroup.option(
                "--dataset-hierarchy",
                type=str,
                default=None,
                help=(
                    "Comma-separated hierarchy of the dataset "
                    "(see http://arcana.readthedocs.io/data_model.html"
                ),
            ),
        ],
    )


def debugging(func):

    return _apply_options(
        func,
        [
            optgroup.group(
                "Debugging options",
                help=(
                    "Options that can be used to help debug the deployment framework"
                ),
            ),
            optgroup.option(
                "--raise-errors/--catch-errors",
                type=bool,
                default=False,
                help="raise exceptions instead of capturing them to suppress call stack",
            ),
            optgroup.option(
                "--keep-running-on-errors/--exit-on-errors",
                type=bool,
                default=False,
                help=(
                    "Keep the the pipeline running in infinite loop on error (will need "
                    "to be manually killed). Can be useful in situations where the "
                    "enclosing container will be removed on completion and you need to "
                    "be able to 'exec' into the container to debug."
                ),
            ),
            optgroup.option(
                "--spec-path",
                type=click.Path(exists=True, path_type=Path),
                default=Path(App.SPEC_PATH),
                help=(
                    "Used to specify a different path to the spec path from the one that is written "
                    "to in the image (typically used in debugging/testing)"
                ),
            ),
        ],
    )


def _apply_options(func, options: list[click.Option]):

    for opt in reversed(options):
        func = opt(func)

    return func



from pathlib import Path
import logging
import cloudpickle as cp
import click
from arcana.core.data.set import Dataset
from arcana.core.utils.misc import set_loggers
from .base import cli

logger = logging.getLogger("arcana")


@cli.group()
def derive():
    pass


@derive.command(
    name="column",
    help="""Derive data for a data sink column and
all prerequisite columns.

DATASET_LOCATOR string containing the nickname of the data store, the ID of the dataset
(e.g. XNAT project ID or file-system directory) and the dataset's name in the
format <store-nickname>//<dataset-id>[@<dataset-name>]

COLUMNS are the names of the sink columns to derive""",
)
@click.argument("dataset_locator")
@click.argument("columns", nargs=-1)
@click.option(
    "--work",
    "-w",
    default=None,
    help=(
        "The location of the directory where the working files "
        "created during the pipeline execution will be stored"
    ),
)
@click.option(
    "--plugin",
    default="cf",
    help=("The Pydra plugin with which to process the workflow"),
)
@click.option(
    "--loglevel",
    type=str,
    default="info",
    help=("The level of detail logging information is presented"),
)
def derive_column(dataset_locator, columns, work, plugin, loglevel):

    logging.basicConfig(level=getattr(logging, loglevel.upper()))

    if work is not None:
        work_dir = Path(work)
        store_cache = work_dir / "store-cache"
        pipeline_cache = work_dir / "pipeline-cache"
    else:
        store_cache = None
        pipeline_cache = None

    dataset = Dataset.load(dataset_locator, cache_dir=store_cache)

    set_loggers(loglevel)

    dataset.derive(*columns, cache_dir=pipeline_cache, plugin=plugin)

    columns_str = "', '".join(columns)
    logger.info(f"Derived data for '{columns_str}' column(s) successfully")


@derive.command(name="output", help="""Derive an output""")
def derive_output():
    raise NotImplementedError


@derive.command(help="""Display the potential derivatives that can be derived""")
def menu():
    raise NotImplementedError


@derive.command(
    name="show-errors",
    help="""Show a Pydra crash report

NODE_WORK_DIR is the directory containing the error pickle file""",
)
@click.argument("node_work_dir")
def show_errors(node_work_dir):
    node_work_dir = Path(node_work_dir)
    files = ["_task.pklz", "_result.pklz", "_error.pklz"]  #
    for fname in files:
        fpath = node_work_dir / fname
        if fpath.exists():
            with open(fpath, "rb") as f:
                contents = cp.load(f)
            click.echo(f"{fname}:")
            if isinstance(contents, dict):
                for k, v in contents.items():
                    if k == "error message":
                        click.echo(f"{k}:\n" + "".join(v))
                    else:
                        click.echo(f"{k}: {v}")
            else:
                click.echo(contents)


@derive.command(
    name="ignore-diff",
    help="""Ignore difference between provenance of previously generated derivative
and new parameterisation""",
)
def ignore_diff():
    raise NotImplementedError


if __name__ == "__main__":
    from click.testing import CliRunner

    runner = CliRunner()
    runner.invoke(
        show_errors,
        [
            "/Users/tclose/Downloads/892d1907-fe2b-40ac-9b77-a3f2ee21ca76/pipeline-cache/Workflow_f3a2bb7474848840aec86fd87e88f5938217fae5976fc699bfc993d95d48a3b8"
        ],
        catch_exceptions=False,
    )



from __future__ import annotations
import os
from pathlib import Path
import typing as ty
from itertools import chain
from copy import copy
import logging
import shutil
import attrs
from attrs.converters import optional
from arcana.core.exceptions import (
    ArcanaUsageError,
    ArcanaFileFormatError,
)
from .base import FileGroup, absolute_path, absolute_paths_dict


logger = logging.getLogger("arcana")


@attrs.define
class BaseFile(FileGroup):

    is_dir = False

    def set_fs_paths(self, fs_paths: ty.List[Path]):
        self._check_paths_exist(fs_paths)
        fs_path = absolute_path(self.matches_ext(*fs_paths))
        self.exists = True
        self.fs_path = fs_path

    def all_file_paths(self):
        """The paths of all nested files within the file-group"""
        if self.fs_path is None:
            raise ArcanaUsageError(
                f"Attempting to access file paths of {self} before they are set"
            )
        return self.fs_paths

    def copy_to(self, fs_path: str or Path, symlink: bool = False):
        """Copies the file-group to the new path, with auxiliary files saved
        alongside the primary-file path.

        Parameters
        ----------
        path : str
            Path to save the file-group to excluding file extensions
        symlink : bool
            Use symbolic links instead of copying files to new location

        Returns
        -------
        BaseFile
            A copy of the file object at the new file system path
        """
        if symlink:
            copy_file = os.symlink
        else:
            copy_file = shutil.copyfile
        dest_path = Path(str(fs_path) + "." + self.ext)
        copy_file(self.fs_path, dest_path)
        cpy = copy(self)
        cpy.set_fs_paths([dest_path])
        return cpy

    @classmethod
    def copy_ext(cls, old_path, new_path):
        """Copy extension from the old path to the new path, ensuring that all
        of the extension is used (e.g. 'nii.gz' instead of 'gz')

        Parameters
        ----------
        old_path: Path or str
            The path from which to copy the extension from
        new_path: Path or str
            The path to append the extension to

        Returns
        -------
        Path
            The new path with the copied extension
        """
        if not cls.matches_ext(old_path):
            raise ArcanaFileFormatError(
                f"Extension of old path ('{str(old_path)}') does not match that "
                f"of file, '{cls.ext}'"
            )
        suffix = "." + cls.ext if cls.ext is not None else old_path.suffix
        return Path(new_path).with_suffix(suffix)

    @classmethod
    def all_exts(cls):
        return [cls.ext]


@attrs.define
class WithSideCars(BaseFile):
    """Base class for file-groups with a primary file and several header or
    side car files
    """

    side_cars: ty.Dict[str, str] = attrs.field(converter=optional(absolute_paths_dict))

    @side_cars.default
    def default_side_cars(self):
        if self.fs_path is None:
            return {}
        return self.default_side_car_paths(self.fs_path)

    @side_cars.validator
    def validate_side_cars(self, _, side_cars):
        if side_cars:
            if self.fs_path is None:
                raise ArcanaUsageError(
                    "Auxiliary files can only be provided to a FileGroup "
                    f"of '{self.path}' ({side_cars}) if the local path is "
                    "as well"
                )
            if set(self.side_car_exts) != set(side_cars.keys()):
                raise ArcanaUsageError(
                    "Keys of provided auxiliary files ('{}') don't match "
                    "format ('{}')".format(
                        "', '".join(side_cars.keys()), "', '".join(self.side_car_exts)
                    )
                )
            missing_side_cars = [(n, f) for n, f in side_cars.items() if not f.exists()]
            if missing_side_cars:
                msg = (
                    f"Attempting to set paths of auxiliary files for {self} "
                    "that don't exist: "
                )
                for name, fpath in missing_side_cars:
                    if fpath.parent.exists():
                        info = "neighbouring files: " + ", ".join(
                            p.name for p in fpath.parent.iterdir()
                        )
                    else:
                        info = "parent directory doesn't exist"
                    msg += f"\n    {name}: {str(fpath)} - {info}"
                raise ArcanaUsageError(msg)

    @classmethod
    def fs_names(cls):
        """Return names for each top-level file-system path in the file group,
        used when generating Pydra task interfaces.

        Returns
        -------
        tuple[str]
            sequence of names for top-level file-system paths in the file group"""
        return super().fs_names() + cls.side_car_exts

    def set_fs_paths(self, paths: ty.List[Path]):
        super().set_fs_paths(paths)
        to_assign = set(Path(p) for p in paths)
        to_assign.remove(self.fs_path)
        # Begin with default side_car paths and override if provided
        default_side_cars = self.default_side_car_paths(self.fs_path)
        for sc_ext in self.side_car_exts:
            try:
                matched = self.side_cars[sc_ext] = absolute_path(
                    self.matches_ext(*paths, ext=sc_ext)
                )
            except ArcanaFileFormatError:
                self.side_cars[sc_ext] = default_side_cars[sc_ext]
            else:
                to_assign.remove(matched)

    @property
    def fs_paths(self):
        return chain(super().fs_paths, self.side_cars.values())

    def side_car(self, name):
        return self.side_cars[name]

    def copy_to(self, fs_path: str or Path, symlink: bool = False):
        """Copies the file-group to the new path, with auxiliary files saved
        alongside the primary-file path.

        Parameters
        ----------
        fs_path : str or Path
            Path to save the file-group to excluding file extensions
        symlink : bool
            Use symbolic links instead of copying files to new location
        """
        if symlink:
            copy_file = os.symlink
        else:
            copy_file = shutil.copyfile
        dest_path = Path(str(fs_path) + "." + self.ext)
        copy_file(self.fs_path, dest_path)
        dest_side_cars = self.default_side_car_paths(dest_path)
        for sc_ext, sc_path in self.side_cars.items():
            copy_file(sc_path, dest_side_cars[sc_ext])
        cpy = copy(self)
        cpy.set_fs_paths([dest_path] + list(dest_side_cars.values()))
        return cpy

    @classmethod
    def default_side_car_paths(cls, primary_path):
        """
        Get the default paths for auxiliary files relative to the path of the
        primary file, i.e. the same name as the primary path with a different
        extension

        Parameters
        ----------
        primary_path : str
            Path to the primary file in the file_group

        Returns
        -------
        aux_paths : ty.Dict[str, str]
            A dictionary of auxiliary file names and default paths
        """

        return {
            e: Path(str(primary_path)[: -len(cls.ext)] + e) for e in cls.side_car_exts
        }

    @classmethod
    def copy_ext(cls, old_path, new_path):
        """Copy extension from the old path to the new path, ensuring that all
        of the extension is used (e.g. 'nii.gz' instead of 'gz'). If the old
        path extension doesn't match the primary path, the methods loops through
        all side-car extensions and selects the longest matching.

        Parameters
        ----------
        old_path: Path or str
            The path from which to copy the extension from
        new_path: Path or str
            The path to append the extension to

        Returns
        -------
        Path
            The new path with the copied extension
        """
        try:
            # Check to see if the path it matches the primary path extension
            return super().copy_ext(old_path, new_path)
        except ArcanaFileFormatError:
            pass
        matches = []
        for ext in cls.side_car_exts:
            try:
                cls.matches_ext(old_path, ext=ext)
            except ArcanaFileFormatError:
                pass
            else:
                matches.append(ext)
        if not matches:
            sc_exts_str = "', '".join(cls.side_car_exts)
            raise ArcanaFileFormatError(
                f"Extension of old path ('{str(old_path)}') does not match any "
                f" in {cls}: '{cls.ext}', {sc_exts_str}"
            )
        longest_match = max(matches, key=len)
        return Path(new_path).with_suffix("." + longest_match)

    def generalise_checksum_keys(
        self, checksums: ty.Dict[str, str], base_path: Path = None
    ):
        """Generalises the paths used for the file paths in a checksum dictionary
        so that they are the same irrespective of that the top-level file-system
        paths are

        Parameters
        ----------
        checksums: dict[str, str]
            The checksum dict mapping relative file paths to checksums

        Returns
        -------
        dict[str, str]
            The checksum dict with file paths generalised"""
        if base_path is None:
            base_path = self.fs_path
        generalised = {}
        fs_name_dict = {
            self.matches_ext(*checksums.keys(), ext=e): e for e in self.side_car_exts
        }
        mapped_exts = list(fs_name_dict.values())
        duplicates = set([e for e in mapped_exts if mapped_exts.count(e) > 1])
        if duplicates:
            raise ArcanaUsageError(
                f"Multiple files with same extensions found in {self}: "
                + ", ".join(str(k) for k in checksums.keys())
            )
        for key, chksum in checksums.items():
            try:
                rel_key = fs_name_dict[str(key)]
            except KeyError:
                try:
                    rel_key = Path(key).relative_to(base_path)
                except ValueError:
                    continue  # skip these files
            generalised[str(rel_key)] = chksum
        return generalised

    @classmethod
    def all_exts(cls):
        return [cls.ext] + list(cls.side_car_exts)



from __future__ import annotations
import shutil
import re
from copy import copy
import tempfile
import json
import logging
from pathlib import Path
import typing as ty
import sys
from collections import defaultdict
import attrs
from attrs.converters import default_if_none
import pydra.engine.task
from pydra.engine.core import TaskBase
from arcana.core.utils.serialize import (
    ObjectListConverter,
    ClassResolver,
)
from arcana.core.utils.misc import show_workflow_errors
from arcana.core.data.row import DataRow
from arcana.core.data.set import Dataset
from arcana.core.data.store import DataStore
from arcana.core.data.space import DataSpace
from arcana.core.exceptions import ArcanaUsageError
from .components import CommandInput, CommandOutput, CommandParameter


if ty.TYPE_CHECKING:
    from ..image import App


logger = logging.getLogger("arcana")


@attrs.define(kw_only=True)
class ContainerCommand:
    """A definition of a command to be run within a container. A command wraps up a
    task or workflow to provide/configure a UI for convenient launching.

    Parameters
    ----------
    task : pydra.engine.task.TaskBase or str
        the task to run or the location of the class
    row_frequency: DataSpace, optional
        the frequency that the command operates on
    inputs: list[CommandInput]
        inputs of the command
    outputs: list[CommandOutput]
        outputs of the command
    parameters: list[CommandParameter]
        parameters of the command
    configuration: dict[str, ty.Any]
        constant values used to configure the task/workflow
    image: App
        back-reference to the image the command is installed in
    """

    STORE_TYPE = "file"
    DATA_SPACE = None

    task: pydra.engine.task.TaskBase = attrs.field(
        converter=ClassResolver(TaskBase, alternative_types=[ty.Callable])
    )
    row_frequency: DataSpace = None
    inputs: list[CommandInput] = attrs.field(
        factory=list,
        converter=ObjectListConverter(CommandInput),
        metadata={"serializer": ObjectListConverter.asdict},
    )
    outputs: list[CommandOutput] = attrs.field(
        factory=list,
        converter=ObjectListConverter(CommandOutput),
        metadata={"serializer": ObjectListConverter.asdict},
    )
    parameters: list[CommandParameter] = attrs.field(
        factory=list,
        converter=ObjectListConverter(CommandParameter),
        metadata={"serializer": ObjectListConverter.asdict},
    )
    configuration: dict[str, ty.Any] = attrs.field(
        factory=dict, converter=default_if_none(dict)
    )
    image: App = None

    def __attrs_post_init__(self):
        if isinstance(self.row_frequency, DataSpace):
            pass
        elif isinstance(self.row_frequency, str):
            try:
                self.row_frequency = DataSpace.fromstr(self.row_frequency)
            except ValueError:
                if self.DATA_SPACE:
                    self.row_frequency = self.DATA_SPACE[self.row_frequency]
                else:
                    raise ValueError(
                        f"'{self.row_frequency}' cannot be resolved to a data space, "
                        "needs to be of form <data-space-enum>[<row-frequency-name>]"
                    )
        elif self.DATA_SPACE:
            self.row_frequency = self.DATA_SPACE.default()
        else:
            raise ValueError(
                f"Value for row_frequency must be provided to {type(self).__name__}.__init__ "
                "because it doesn't have a defined DATA_SPACE class attribute"
            )

    @property
    def name(self):
        if self.image is None:
            raise RuntimeError(
                f"Cannot access name of unbound container commands {self}"
            )
        return self.image.name

    @property
    def data_space(self):
        return type(self.row_frequency)

    def activate_conda_cmd(self):
        """Generate the preamble to a command line that activates the conda environment

        Returns
        -------
        str
            part of a command line, which activates the conda environment
        """

        return f"conda run --no-capture-output -n {self.image.CONDA_ENV} "  # activate conda

    def configuration_args(self):

        # Set up fixed arguments used to configure the workflow at initialisation
        cmd_args = []
        if self.configuration is not None:
            for cname, cvalue in self.configuration.items():
                cvalue_json = json.dumps(cvalue)
                cmd_args.append(f"--configuration {cname} '{cvalue_json}' ")

        return cmd_args

    def license_args(self):
        cmd_args = []
        for lic_name, lic in self.image.licenses.items():
            if lic.source is None:
                cmd_args.append(f"--download-license {lic_name} {lic.destination}")
        return cmd_args

    def execute(
        self,
        dataset_locator: str,
        input_values: dict[str, str] = None,
        output_values: dict[str, str] = None,
        parameter_values: dict[str, ty.Any] = None,
        work_dir: Path = None,
        ids: list[str] = None,
        single_row: str = None,
        dataset_hierarchy: str = None,
        dataset_name: str = None,
        overwrite: bool = False,
        loglevel: str = "warning",
        plugin: str = None,
        export_work: Path = False,
        raise_errors: bool = False,
        keep_running_on_errors=False,
        pipeline_name: str = None,
    ):
        """Runs the command within the entrypoint of the container image.

        Performs a number of steps in one long pipeline that would typically be done
        in separate command calls when running manually, i.e.:

            * Loads a dataset, creating if it doesn't exist
            * create input and output columns if they don't exist
            * applies the pipeline to the dataset
            * runs the pipeline

        Parameters
        ----------
        dataset : Dataset
            dataset ID str (<store-nickname>//<dataset-id>:<dataset-name>)
        input_values : dict[str, str]
            values passed to the inputs of the command
        output_values : dict[str, str]
            values passed to the outputs of the command
        parameter_values : dict[str, ty.Any]
            values passed to the parameters of the command
        store_cache_dir : Path
            cache path used to download data from the store to the working node (if necessary)
        pipeline_cache_dir : Path
            cache path created when running the pipelines
        plugin : str
            Pydra plugin used to execute the pipeline
        ids : list[str]
            IDs of the dataset rows to run the pipeline over
        overwrite : bool, optional
            overwrite existing outputs
        export_work : Path
            export work directory to an alternate location after the workflow is run
            (e.g. for forensics)
        raise_errors : bool
            raise errors instead of capturing and logging (for debugging)
        pipeline_name : str
            the name to give to the pipeline, defaults to the name of the command image
        """

        if type(export_work) is bytes:
            export_work = Path(export_work.decode("utf-8"))

        if loglevel != "none":
            logging.basicConfig(
                stream=sys.stdout, level=getattr(logging, loglevel.upper())
            )

        if work_dir is None:
            work_dir = tempfile.mkdtemp()

        if pipeline_name is None:
            pipeline_name = self.name

        work_dir = Path(work_dir)
        work_dir.mkdir(parents=True, exist_ok=True)

        store_cache_dir = work_dir / "store-cache"
        pipeline_cache_dir = work_dir / "pydra"

        dataset = self.load_dataset(
            dataset_locator, store_cache_dir, dataset_hierarchy, dataset_name
        )

        if single_row is not None:
            # Adds a single row to the dataset (i.e. skips a full scan)
            dataset.add_leaf(single_row.split(","))

        # Install required software licenses from store into container
        if self.image is not None:
            dataset.download_licenses(
                [lic for lic in self.image.licenses if not lic.store_in_image]
            )

        input_values = dict(input_values) if input_values else {}
        output_values = dict(output_values) if output_values else {}
        parameter_values = dict(parameter_values) if parameter_values else {}

        input_configs = []
        converter_args = {}  # Arguments passed to converter
        for inpt in self.inputs:
            if not input_values[inpt.name] and inpt.datatype != DataRow:
                logger.warning(
                    f"Skipping '{inpt.name}' source column as no input was provided"
                )
                continue
            if inpt.datatype is DataRow:
                logger.info(
                    f"No column added for '{inpt.name}' column as it uses built-in "
                    "type `arcana.core.data.row.DataRow`"
                )
                continue
            path, qualifiers = self.extract_qualifiers_from_path(
                input_values[inpt.name]
            )
            source_kwargs = qualifiers.pop("criteria", {})
            converter_args[inpt.name] = qualifiers.pop("converter", {})
            if qualifiers:
                raise ArcanaUsageError(
                    "Unrecognised qualifier namespaces extracted from path for "
                    f"{inpt.name} (expected ['criteria', 'converter']): {qualifiers}"
                )
            if inpt.name in dataset.columns:
                column = dataset[inpt.name]
                logger.info(f"Found existing source column {column}")
            else:
                logger.info(f"Adding new source column '{inpt.name}'")
                dataset.add_source(
                    name=inpt.name,
                    datatype=inpt.default_column.datatype,
                    path=path,
                    is_regex=True,
                    **source_kwargs,
                )
            if input_config := inpt.config_dict:
                input_configs.append(input_config)

        output_configs = []
        for output in self.outputs:
            path, qualifiers = self.extract_qualifiers_from_path(
                output_values.get(output.name, output.name)
            )
            converter_args[output.name] = qualifiers.pop("converter", {})
            if qualifiers:
                raise ArcanaUsageError(
                    "Unrecognised qualifier namespaces extracted from path for "
                    f"{output.name} (expected ['criteria', 'converter']): {qualifiers}"
                )
            if output.name in dataset.columns:
                column = dataset[output.name]
                if not column.is_sink:
                    raise ArcanaUsageError(
                        "Output column name '{output.name}' shadows existing source column"
                    )
                logger.info(f"Found existing sink column {column}")
            else:
                logger.info(f"Adding new source column '{output.name}'")
                dataset.add_sink(
                    name=output.name,
                    datatype=output.default_column.datatype,
                    path=path,
                )
            if output_config := output.config_dict:
                output_configs.append(output_config)

        kwargs = copy(self.configuration)

        param_configs = []
        for param in self.parameters:
            param_value = parameter_values.get(param.name, None)
            logger.info(
                "Parameter %s (type %s) passed value %s",
                param.name,
                param.datatype,
                param_value,
            )
            if param_value == "" and param.datatype is not str:
                param_value = None
                logger.info(
                    "Non-string parameter '%s' passed empty string, setting to NOTHING",
                    param.name,
                )
            if param_value is None:
                if param.default is None:
                    raise RuntimeError(
                        f"A value must be provided to required '{param.name}' parameter"
                    )
                param_value = param.default
                logger.info("Using default value for %s, %s", param.name, param_value)

            # Convert parameter to parameter type
            try:
                param_value = param.datatype(param_value)
            except ValueError:
                raise ValueError(
                    f"Could not convert value passed to '{param.name}' parameter, "
                    f"{param_value}, into {param.datatype}"
                )
            kwargs[param.field] = param_value
            if param_config := param.config_dict:
                param_configs.append(param_config)

        if "name" not in kwargs:
            kwargs["name"] = "pipeline_task"

        if input_configs:
            kwargs["inputs"] = input_configs
        if output_configs:
            kwargs["outputs"] = output_configs
        if param_configs:
            kwargs["parameters"] = param_configs

        task = self.task(**kwargs)

        if pipeline_name in dataset.pipelines and not overwrite:
            pipeline = dataset.pipelines[self.name]
            if task != pipeline.workflow:
                raise RuntimeError(
                    f"A pipeline named '{self.name}' has already been applied to "
                    "which differs from one specified. Please use '--overwrite' option "
                    "if this is intentional"
                )
        else:
            pipeline = dataset.apply_pipeline(
                pipeline_name,
                task,
                inputs=self.inputs,
                outputs=self.outputs,
                row_frequency=self.row_frequency,
                overwrite=overwrite,
                converter_args=converter_args,
            )

        # Instantiate the Pydra workflow
        wf = pipeline(cache_dir=pipeline_cache_dir)

        if ids is not None:
            ids = ids.split(",")

        # execute the workflow
        try:
            result = wf(ids=ids, plugin=plugin)
        except Exception:
            msg = show_workflow_errors(
                pipeline_cache_dir, omit_nodes=["per_node", wf.name]
            )
            logger.error(
                "Pipeline failed with errors for the following nodes:\n\n%s", msg
            )
            if raise_errors or not msg:
                raise
            else:
                errors = True
        else:
            logger.info(
                "Pipeline '%s' ran successfully for the following data rows:\n%s",
                pipeline_name,
                "\n".join(result.output.processed),
            )
            errors = False
        finally:
            if export_work:
                logger.info("Exporting work directory to '%s'", export_work)
                export_work.mkdir(parents=True, exist_ok=True)
                shutil.copytree(pipeline_cache_dir, export_work / "pydra")

        # Abort at the end after the working directory can be copied back to the
        # host so that XNAT knows there was an error
        if errors:
            if keep_running_on_errors:
                while True:
                    pass
            else:
                sys.exit(1)

    @classmethod
    def extract_qualifiers_from_path(cls, user_input: str):
        """Extracts out "qualifiers" from the user-inputted paths. These are
        in the form 'path ns1.arg1=val1 ns1.arg2=val2, ns2.arg1=val3...

        Parameters
        ----------
        col_name : str
            name of the column the
        user_input : str
            The path expression + qualifying keyword args to extract

        Returns
        -------
        path : str
            the path expression stripped of qualifiers
        qualifiers : defaultdict[dict]
            the extracted qualifiers
        """
        qualifiers = defaultdict(dict)
        if "=" in user_input:  # Treat user input as containing qualifiers
            parts = re.findall(r'(?:[^\s"]|"(?:\\.|[^"])*")+', user_input)
            path = parts[0].strip('"')
            for part in parts[1:]:
                try:
                    full_name, val = part.split("=", maxsplit=1)
                except ValueError as e:
                    e.args = ((e.args[0] + f" attempting to split '{part}' by '='"),)
                    raise e
                try:
                    ns, name = full_name.split(".", maxsplit=1)
                except ValueError as e:
                    e.args = (
                        (e.args[0] + f" attempting to split '{full_name}' by '.'"),
                    )
                    raise e
                try:
                    val = json.loads(val)
                except json.JSONDecodeError:
                    pass
                qualifiers[ns][name] = val
        else:
            path = user_input
        return path, qualifiers

    def load_dataset(
        self,
        dataset_locator: str,
        cache_dir: Path,
        dataset_hierarchy: str,
        dataset_name: str,
    ):
        """Loads a dataset from within an image, to be used in image entrypoints

        Parameters
        ----------
        dataset_locator : str
            dataset ID str
        cache_dir : Path
            the directory to use for the store cache
        dataset_hierarchy : str, optional
            the hierarchy of the dataset
        dataset_name : str
            overwrite dataset name loaded from ID str

        Returns
        -------
        _type_
            _description_
        """
        try:
            dataset = Dataset.load(dataset_locator)
        except KeyError:

            store_name, id, name = Dataset.parse_id_str(dataset_locator)

            if dataset_name is not None:
                name = dataset_name

            store = DataStore.load(store_name, cache_dir=cache_dir)

            if dataset_hierarchy is None:
                hierarchy = self.data_space.default().span()
            else:
                hierarchy = dataset_hierarchy.split(",")

            try:
                dataset = store.load_dataset(
                    id, name
                )  # FIXME: Does this need to be here or this covered by L253??
            except KeyError:
                dataset = store.new_dataset(
                    id, hierarchy=hierarchy, space=self.data_space
                )
        return dataset



import click
import typing as ty
from arcana.core.data.set import Dataset
from pydra.engine.core import TaskBase
from arcana.core.utils.serialize import ClassResolver, parse_value
from arcana.core.data.type.base import DataType
from .base import cli


@cli.group()
def apply():
    pass


@apply.command(
    name="pipeline",
    help="""Apply a Pydra workflow to a dataset as a pipeline between
two columns

DATASET_LOCATOR string containing the nickname of the data store, the ID of the dataset
(e.g. XNAT project ID or file-system directory) and the dataset's name in the
format <store-nickname>//<dataset-id>[@<dataset-name>]

PIPELINE_NAME is the name of the pipeline

WORKFLOW_LOCATION is the location to a Pydra workflow on the Python system path,
<MODULE>:<WORKFLOW>""",
)
@click.argument("dataset_locator")
@click.argument("pipeline_name")
@click.argument("workflow_location")
@click.option(
    "--input",
    "-i",
    nargs=3,
    default=(),
    metavar="<col-name> <pydra-field> <required-datatype>",
    multiple=True,
    type=str,
    help=(
        "the link between a column and an input of the workflow. "
        "The required format is the location (<module-path>:<class>) of the format "
        "expected by the workflow"
    ),
)
@click.option(
    "--output",
    "-o",
    nargs=3,
    default=(),
    metavar="<col-name> <pydra-field> <produced-datatype>",
    multiple=True,
    type=str,
    help=(
        "the link between an output of the workflow and a sink column. "
        "The produced datatype is the location (<module-path>:<class>) of the datatype "
        "produced by the workflow"
    ),
)
@click.option(
    "--parameter",
    "-p",
    nargs=2,
    default=(),
    metavar="<name> <value>",
    multiple=True,
    type=str,
    help=("a fixed parameter of the workflow to set when applying it"),
)
@click.option(
    "--source",
    "-s",
    nargs=3,
    default=(),
    metavar="<col-name> <pydra-field> <required-datatype>",
    multiple=True,
    type=str,
    help=(
        "add a source to the dataset and link it to an input of the workflow "
        "in a single step. The source column must be able to be specified by its "
        "path alone and be already in the datatype required by the workflow"
    ),
)
@click.option(
    "--sink",
    "-k",
    nargs=3,
    default=(),
    metavar="<col-name> <pydra-field> <produced-datatype>",
    multiple=True,
    type=str,
    help=(
        "add a sink to the dataset and link it to an output of the workflow "
        "in a single step. The sink column be in the same datatype as produced "
        "by the workflow"
    ),
)
@click.option(
    "--row-frequency",
    "-f",
    default=None,
    type=str,
    help=(
        "the row-frequency of the rows the pipeline will be executed over, i.e. "
        "will it be run once per-session, per-subject or per whole dataset, "
        "by default the highest row-frequency rows (e.g. per-session)"
    ),
)
@click.option(
    "--overwrite/--no-overwrite",
    default=False,
    help=("whether to overwrite previous pipelines"),
)
def apply_pipeline(
    dataset_locator,
    pipeline_name,
    workflow_location,
    input,
    output,
    parameter,
    source,
    sink,
    row_frequency,
    overwrite,
):

    dataset = Dataset.load(dataset_locator)
    workflow = ClassResolver(TaskBase, alternative_types=[ty.Callable])(
        workflow_location
    )(name="workflow", **{n: parse_value(v) for n, v in parameter})

    inputs = parse_col_option(input)
    outputs = parse_col_option(output)
    sources = parse_col_option(source)
    sinks = parse_col_option(sink)

    for col_name, field, datatype in sources:
        dataset.add_source(col_name, datatype)
        inputs.append((col_name, field, datatype))

    for col_name, field, datatype in sinks:
        dataset.add_sink(col_name, datatype)
        outputs.append((col_name, field, datatype))

    dataset.apply_pipeline(
        pipeline_name,
        workflow,
        inputs,
        outputs,
        row_frequency=row_frequency,
        overwrite=overwrite,
    )

    dataset.save()


@apply.command(name="analysis", help="""Applies an analysis class to a dataset""")
def apply_analysis():
    raise NotImplementedError


@apply.command(name="bids-app", help="Apply a BIDS app to a dataset as a pipeline")
@click.option(
    "--container",
    nargs=2,
    default=None,
    metavar="<engine-tag>",
    help=(
        "The container engine ('docker'|'singularity') and the image"
        " to run the app in"
    ),
)
@click.option(
    "--virtualisation",
    default="none",
    type=click.Choice(["docker", "singularity", "none"], case_sensitive=False),
    help=(
        "The virtualisation method to run with the task with (only "
        "applicable to BIDS app tasks)"
    ),
)
def apply_bids_app():
    raise NotImplementedError


def parse_col_option(option):
    return [(c, p, ClassResolver(DataType)(f)) for c, p, f in option]



class ArcanaException(Exception):
    @property
    def msg(self):
        return self.args[0]

    @msg.setter
    def msg(self, msg):
        self.args = (msg,) + self.args[1:]


class ArcanaError(ArcanaException):
    pass


class ArcanaRuntimeError(ArcanaError):
    pass


class ArcanaNotBoundToAnalysisError(ArcanaError):
    pass


class ArcanaVersionError(ArcanaError):
    pass


class ArcanaRequirementNotFoundError(ArcanaVersionError):
    pass


class ArcanaVersionNotDetectableError(ArcanaVersionError):
    pass


class ArcanaEnvModuleNotLoadedError(ArcanaError):
    pass


class ArcanaMissingInputError(ArcanaException):
    pass


class ArcanaProtectedOutputConflictError(ArcanaError):
    pass


class ArcanaCantPickleAnalysisError(ArcanaError):
    pass


class ArcanaRepositoryError(ArcanaError):
    pass


class ArcanaUsageError(ArcanaError):
    pass


class ArcanaCacheError(ArcanaError):
    pass


class ArcanaDesignError(ArcanaError):
    pass


class NamedArcanaError(ArcanaError):
    def __init__(self, name, msg):
        super(NamedArcanaError, self).__init__(msg)
        self.name = name


class ArcanaNameError(NamedArcanaError):
    pass


class ArcanaWrongFrequencyError(NamedArcanaError):
    pass


class ArcanaIndexError(ArcanaError):
    def __init__(self, index, msg):
        super(ArcanaIndexError, self).__init__(msg)
        self.index = index


class ArcanaDataNotDerivedYetError(NamedArcanaError, ArcanaDesignError):
    pass


class ArcanaDataMatchError(ArcanaUsageError):
    pass


class ArcanaPipelinesStackError(ArcanaError):
    pass


class ArcanaMissingDataException(ArcanaPipelinesStackError):
    pass


class ArcanaOutputNotProducedException(ArcanaPipelinesStackError):
    """
    Raised when a given spec is not produced due to switches and inputs
    provided to the analysis
    """


class ArcanaInsufficientRepoDepthError(ArcanaError):
    pass


class ArcanaFileFormatError(ArcanaError):
    pass


class ArcanaLicenseNotFoundError(ArcanaNameError):
    pass


class ArcanaUnresolvableFormatException(ArcanaException):
    pass


class ArcanaFileGroupNotCachedException(ArcanaException):
    pass


class NoMatchingPipelineException(ArcanaException):
    pass


class ArcanaModulesError(ArcanaError):
    pass


class ArcanaModulesNotInstalledException(ArcanaException):
    pass


class ArcanaJobSubmittedException(ArcanaException):
    """
    Signifies that a pipeline has been submitted to a scheduler and
    a return value won't be returned.
    """


class ArcanaNoRunRequiredException(ArcanaException):
    """
    Used to signify when a pipeline doesn't need to be run as all
    required outputs are already present in the store
    """


class ArcanaFileFormatClashError(ArcanaError):
    """
    Used when two mismatching data formats are registered with the same
    name or extension
    """


class ArcanaFormatConversionError(ArcanaError):
    "No converters exist between formats"


class ArcanaConverterNotAvailableError(ArcanaError):
    "The converter required to convert between formats is not"
    "available"


class ArcanaReprocessException(ArcanaException):
    pass


class ArcanaWrongRepositoryError(ArcanaError):
    pass


class ArcanaIvalidParameterError(ArcanaError):
    pass


class ArcanaRequirementVersionsError(ArcanaError):
    pass


class ArcanaXnatCommandError(ArcanaRepositoryError):
    """
    Error in the command file used to access an XNAT repository via the XNAT
    container service.
    """


class ArcanaUriAlreadySetException(ArcanaException):
    """Raised when attempting to set the URI of an item is already set"""


class ArcanaDataTreeConstructionError(ArcanaError):
    "Error in constructing data tree by store find_rows method"


class ArcanaBadlyFormattedIDError(ArcanaDataTreeConstructionError):
    "Error attempting to extract an ID from a tree path using a user provided regex"


class ArcanaWrongDataSpaceError(ArcanaError):
    "Provided row_frequency is not a valid member of the dataset's dimensions"


class ArcanaNoDirectXnatMountException(ArcanaException):
    "Raised when attemptint to access a file-system mount for a row that hasn't been mounted directly"
    pass


class ArcanaEmptyDatasetError(ArcanaException):
    pass


class ArcanaBuildError(ArcanaError):
    pass



from __future__ import annotations
from typing import Iterable, Union, List, Union, Dict, Tuple
import shutil
import re
import yaml
import tempfile
from pathlib import Path
import attrs
import pytest
import docker
from arcana.core.cli.deploy import (
    make_app,
    make_docs,
)
from arcana.core.utils.testing import show_cli_trace
from arcana.core.exceptions import ArcanaBuildError


def test_deploy_make_app_cli(command_spec, cli_runner, work_dir):

    DOCKER_ORG = "testorg"
    DOCKER_REGISTRY = "test.registry.org"
    IMAGE_GROUP_NAME = "testpkg"

    concatenate_spec = {
        "command": command_spec,
        "version": "1.0",
        "build_iteration": "1",
        "packages": {
            "system": ["vim"],  # just to test it out
            "pip": {"click": None},  # just to test it out
        },
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "info_url": "http://concatenate.readthefakedocs.io",
        "description": "a test image spec",
    }

    build_dir = work_dir / "build"
    build_dir.mkdir()
    spec_path = work_dir / DOCKER_ORG
    sub_dir = spec_path / IMAGE_GROUP_NAME
    sub_dir.mkdir(parents=True)
    with open(sub_dir / "concatenate.yml", "w") as f:
        yaml.dump(concatenate_spec, f)

    result = cli_runner(
        make_app,
        [
            str(spec_path),
            "common:App",
            "--build-dir",
            str(build_dir),
            "--registry",
            DOCKER_REGISTRY,
            "--loglevel",
            "warning",
            "--use-local-packages",
            "--install-extras",
            "test",
            "--raise-errors",
            "--use-test-config",
            "--dont-check-registry",
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    tag = result.output.strip().splitlines()[-1]
    assert tag == f"{DOCKER_REGISTRY}/{DOCKER_ORG}/{IMAGE_GROUP_NAME}.concatenate:1.0-1"

    # Clean up the built image
    dc = docker.from_env()
    dc.images.remove(tag)


def test_deploy_remake_app_cli(command_spec, docker_registry, cli_runner, run_prefix):
    """Tests the check to see whether"""

    IMAGE_GROUP_NAME = "testpkg-rebuild" + run_prefix

    def build_spec(spec, **kwargs):
        work_dir = Path(tempfile.mkdtemp())
        build_dir = work_dir / "build"
        build_dir.mkdir()
        spec_path = work_dir / "testorg"
        sub_dir = spec_path / IMAGE_GROUP_NAME
        sub_dir.mkdir(parents=True)
        with open(sub_dir / "concatenate.yml", "w") as f:
            yaml.dump(spec, f)

        result = cli_runner(
            make_app,
            [
                str(spec_path),
                "common:App",
                "--build-dir",
                str(build_dir),
                "--registry",
                docker_registry,
                "--loglevel",
                "warning",
                "--use-local-packages",
                "--install-extras",
                "test",
                "--raise-errors",
                "--check-registry",
                "--use-test-config",
            ],
            **kwargs,
        )
        return result

    concatenate_spec = {
        "command": command_spec,
        "version": "1.0",
        "build_iteration": "1",
        "packages": {"system": ["vim"]},
        "description": "a test image",
        "name": "test_deploy_rebuild_cli",
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "info_url": "http://concatenate.readthefakedocs.io",
    }

    # Build a basic image
    result = build_spec(concatenate_spec)
    assert result.exit_code == 0, show_cli_trace(result)
    assert result.output
    tag = result.output.strip().splitlines()[-1]
    try:
        dc = docker.from_env()
        dc.api.push(tag)

        # FIXME: Need to ensure that logs are captured properly then we can test this
        # result = build_spec(concatenate_spec)
        # assert "Skipping" in result.output

        # Modify the spec so it doesn't match the original that has just been
        # built (but don't increment the version number -> image tag so there
        # is a clash)
        concatenate_spec["packages"] = {"system": ["vim"]}

        with pytest.raises(ArcanaBuildError) as excinfo:
            build_spec(concatenate_spec, catch_exceptions=False)

        assert "doesn't match the one that was used to build the image" in str(
            excinfo.value
        )

        # Increment the version number to avoid the clash
        concatenate_spec["build_iteration"] = "2"

        result = build_spec(concatenate_spec)
        assert result.exit_code == 0, show_cli_trace(result)
        rebuilt_tag = result.output.strip().splitlines()[-1]
        dc.images.remove(rebuilt_tag)
    finally:
        # Clean up the built images
        dc.images.remove(tag)


@attrs.define
class DocsFixture:

    yaml_src: str
    markdown: str
    licenses_to_provide: list[str] = attrs.field(factory=list)


docs_fixtures = {
    "simple": DocsFixture(
        """
version: &version '0.16.1'
authors:
  - name: author_name
    email: author@email.org
base_image:
  name: abc
  tag: *version
  package_manager: apt
info_url: https://example.com
description: >-
  a test of the YAML join functionality
command:
  task: arcana.core.utils.testing.tasks:identity_file
  row_frequency: common:Samples[sample]
  inputs:
    in_file:
      datatype: common:Text
      help_string: the input file
  outputs:
    out_file:
      datatype: common:Text
      help_string: the output file
    """.strip(),
        """
---
source_file: spec.yaml
title: spec
weight: 10

---

## Package Info
|Key|Value|
|---|-----|
|Name|spec|
|App version|0.16.1|
|Build iteration|0|
|Base image|`abc:0.16.1`|
|Maintainer|author_name (author@email.org)|
|Info URL|https://example.com|
|Short description|a test of the YAML join functionality|

a test of the YAML join functionality

## Command
|Key|Value|
|---|-----|
|Task|arcana.core.utils.testing.tasks:identity_file|
|Operates on|sample|
#### Inputs
|Name|Data type|Stored data type default|Description|
|----|---------|------------------------|-----------|
|`in_file`|<span data-toggle="tooltip" data-placement="bottom" title="text" aria-label="text">text (`.txt`)</span>|<span data-toggle="tooltip" data-placement="bottom" title="text" aria-label="text">text (`.txt`)</span>|the input file|

#### Outputs
|Name|Data type|Stored data type default|Description|
|----|---------|------------------------|-----------|
|`out_file`|<span data-toggle="tooltip" data-placement="bottom" title="text" aria-label="text">text (`.txt`)</span>|<span data-toggle="tooltip" data-placement="bottom" title="text" aria-label="text">text (`.txt`)</span>|the output file|

#### Parameters
|Name|Data type|Description|
|----|---------|-----------|
""".strip(),
    ),
    "full": DocsFixture(
        """
version: &version '0.16.1'
build_iteration: '10'
authors:
  - name: author_name
    email: author@email.org
base_image:
  name: abc
  tag: *version
  package_manager: yum
description: a description
long_description: >-
  a longer description
known_issues:
  - url: https://example.com
info_url: https://example.com
packages:
  system:
    vim: 99.1
    git:
  pip:
    - pydra
    - pydra-dcm2niix
  neurodocker:
    dcm2niix: v1.0.20201102
licenses:
  freesurfer:
    destination: /opt/freesurfer/license.txt
    description: >
      license description
    info_url: http://path.to.license.provider.org/licenses
command:
    task: bids:bids_app
    inputs:
      T1w:
        configuration:
          path: anat/T1w
        datatype: medimage:NiftiGzX
        help_string: "T1-weighted anatomical scan"
        default_column:
          datatype: medimage:Dicom
      T2w:
        configuration:
          path: anat/T2w
        datatype: medimage:NiftiGzX
        help_string: "T2-weighted anatomical scan"
        default_column:
          datatype: medimage:Dicom
      fMRI:
        datatype: medimage:NiftiGzX
        help_string: "functional MRI"
        configuration:
          path: func/task-rest_bold
        default_column:
          datatype: medimage:Dicom
    outputs:
      mriqc:
        datatype: common:Directory
        help_string: "MRIQC output directory"
        configuration:
          path: mriqc
    parameters:
      fmriprep_flags:
        field: flags
        datatype: str
        help_string: description of flags param
    row_frequency: medimage:Clinical[session]
    configuration:
      executable: /usr/local/miniconda/bin/mriqc
      dataset: /work/bids-dataset
      app_output_dir: /work/bids-app-output
    """.strip(),
        """
---
source_file: /var/folders/mz/yn83q2fd3s758w1j75d2nnw80000gn/T/tmp47_dxmyq/specs/spec.yaml
title: spec
weight: 10

---

## Package Info
|Key|Value|
|---|-----|
|Name|spec|
|App version|0.16.1|
|Build iteration|10|
|Base image|`abc:0.16.1`|
|Maintainer|author_name (author@email.org)|
|Info URL|https://example.com|
|Short description|a description|
|Known issues|https://example.com|

a longer description

### Required licenses
|Name|URL|Description|
|----|---|-----------|
|freesurfer|`http://path.to.license.provider.org/licenses`|license description|

## Command
|Key|Value|
|---|-----|
|Task|bids:bids_app|
|Operates on|session|
#### Inputs
|Name|Data type|Stored data type default|Description|
|----|---------|------------------------|-----------|
|`T1w`|<span data-toggle="tooltip" data-placement="bottom" title="niftigzx" aria-label="niftigzx">niftigzx (from 'medimage' extension)</span>|<span data-toggle="tooltip" data-placement="bottom" title="dicom" aria-label="dicom">dicom (from 'medimage' extension)</span>|T1-weighted anatomical scan|
|`T2w`|<span data-toggle="tooltip" data-placement="bottom" title="niftigzx" aria-label="niftigzx">niftigzx (from 'medimage' extension)</span>|<span data-toggle="tooltip" data-placement="bottom" title="dicom" aria-label="dicom">dicom (from 'medimage' extension)</span>|T2-weighted anatomical scan|
|`fMRI`|<span data-toggle="tooltip" data-placement="bottom" title="niftigzx" aria-label="niftigzx">niftigzx (from 'medimage' extension)</span>|<span data-toggle="tooltip" data-placement="bottom" title="dicom" aria-label="dicom">dicom (from 'medimage' extension)</span>|functional MRI|

#### Outputs
|Name|Data type|Stored data type default|Description|
|----|---------|------------------------|-----------|
|`mriqc`|<span data-toggle="tooltip" data-placement="bottom" title="directory" aria-label="directory">directory (directory)</span>|<span data-toggle="tooltip" data-placement="bottom" title="directory" aria-label="directory">directory (directory)</span>|MRIQC output directory|

#### Parameters
|Name|Data type|Description|
|----|---------|-----------|
|`fmriprep_flags`|`str`|description of flags param|
""".strip(),
        ["freesurfer"],
    ),
}


@pytest.mark.parametrize("fixture", docs_fixtures.items(), ids=lambda x: x[0])
def test_make_docs_cli(
    cli_runner, run_prefix, work_dir: Path, fixture: Tuple[str, DocsFixture]
):
    fixture_name, fixture_content = fixture

    # TODO handle multiple 'files' in a fixture
    print(f"Processing fixture: {fixture_name!r}")
    output = _make_docs(cli_runner, work_dir, fixture_content.yaml_src)

    strip_source_file_re = re.compile(r"source_file:.*")

    stripped_output = strip_source_file_re.sub("", output)
    stripped_reference = strip_source_file_re.sub("", fixture_content.markdown)

    assert (
        stripped_output == stripped_reference
    ), f"Fixture {fixture_name!r} didn't match output"


def _make_docs(
    cli_runner,
    work_dir: Path,
    docs: Union[str, Dict[str, str]],
    *args,
    flatten: bool = None,
) -> Union[str, Dict[str, str]]:
    out_dir = work_dir / "out"
    specs_dir = work_dir / "specs"
    if specs_dir.exists():
        shutil.rmtree(specs_dir)
    specs_dir.mkdir()

    if type(docs) is str:
        (specs_dir / "spec.yaml").write_text(docs)
    else:
        for name, content in docs.items():
            path = specs_dir / name
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text(content)

    result = cli_runner(
        make_docs,
        [
            specs_dir.as_posix(),
            out_dir.as_posix(),
        ]
        + (["--flatten" if flatten else "--no-flatten"] if flatten is not None else [])
        + list(args),
    )

    assert result.exit_code == 0, show_cli_trace(result)

    if type(docs) is str:
        return (out_dir / "spec.md").read_text().strip()
    else:
        return {
            file.relative_to(out_dir).as_posix(): file.read_text().strip()
            for file in out_dir.glob("*.md")
        }



import zipfile
import tempfile
from pathlib import Path
from arcana.dirtree.data import Text, Zip
from arcana.core.utils.testing.data import make_dataset, TestDataSpace
from conftest import TEST_DATASET_BLUEPRINTS
from arcana.core.utils.testing.tasks import concatenate

# from pydra.tasks.fsl.preprocess.bet import BET
# from fileformats.medimage.data import Dicom, NiftiGz


def test_pipeline(work_dir):
    dataset = make_dataset(
        TEST_DATASET_BLUEPRINTS["concatenate_test"], work_dir / "dataset"
    )

    dataset.add_source("file1", Text)
    dataset.add_source("file2", Text)
    dataset.add_sink("deriv", Text)

    pipeline = dataset.apply_pipeline(
        name="test_pipeline",
        workflow=concatenate(duplicates=2, name="concatenate"),
        inputs=[("file1", "in_file1"), ("file2", "in_file2")],
        outputs=[("deriv", "out_file")],
        row_frequency=TestDataSpace.abcd,
    )

    IDS = ["a0b0c0d0", "a0b0c0d1"]

    workflow = pipeline(cache_dir=work_dir / "pipeline-cache")
    workflow(ids=IDS, plugin="serial")

    dataset.refresh()  # Reset cached values

    for item in dataset["deriv"]:
        with open(item.fs_path) as f:
            contents = f.read()
        assert contents == "\n".join(["file1.txt", "file2.txt"] * 2)


def test_pipeline_with_implicit_conversion(work_dir):
    """Input files are converted from zip to Text, concatenated and then
    written back as zip files into the data store"""
    dataset = make_dataset(
        TEST_DATASET_BLUEPRINTS["concatenate_zip_test"], work_dir / "dataset"
    )

    dataset.add_source("file1", Zip)
    dataset.add_source("file2", Zip)
    dataset.add_sink("deriv", Zip)

    pipeline = dataset.apply_pipeline(
        name="test_pipeline",
        workflow=concatenate(duplicates=2, name="concatenate"),
        inputs=[("file1", "in_file1", Text), ("file2", "in_file2", Text)],
        outputs=[("deriv", "out_file", Text)],
        row_frequency=TestDataSpace.abcd,
    )

    IDS = ["a0b0c0d0", "a0b0c0d1"]

    workflow = pipeline(cache_dir=work_dir / "pipeline-cache")
    workflow(ids=IDS, plugin="serial")

    dataset.refresh()

    for item in dataset["deriv"]:
        tmp_dir = Path(tempfile.mkdtemp())
        with zipfile.ZipFile(item.fs_path) as zfile:
            zfile.extractall(path=tmp_dir)
        with open(tmp_dir / "out_u_file.txt") as f:
            contents = f.read()
        assert contents == "\n".join(["file1.zip", "file2.zip"] * 2)



from __future__ import annotations
from pathlib import Path
import typing as ty
from copy import copy
import hashlib
import logging
from abc import ABCMeta, abstractmethod
import attrs
from attrs.converters import optional
from pydra.engine.core import LazyField, Workflow
from arcana.core.utils.misc import (
    func_task,
    path2varname,
    classproperty,
    CONVERTER_ANNOTATIONS,
    HASH_CHUNK_SIZE,
)
from arcana.core.exceptions import (
    ArcanaUsageError,
    ArcanaNameError,
    ArcanaDataNotDerivedYetError,
    ArcanaFileFormatError,
    ArcanaFormatConversionError,
)
from ...analysis.salience import DataQuality  # @ignore reshadowedImports


if ty.TYPE_CHECKING:
    from arcana.core.data.row import DataRow


def absolute_path(path):
    return Path(path).absolute()


def absolute_paths_dict(dct):
    return {n: absolute_path(p) for n, p in dict(dct).items()}


logger = logging.getLogger("arcana")


@attrs.define
class DataType(metaclass=ABCMeta):
    """
    A representation of an atomic data item within a dataset. Subclassed to define
    different data file/directory formats and fields

    Parameters
    ----------
    name_path : str
        The name_path to the relative location of the file group, i.e. excluding
        information about which row in the data tree it belongs to
    order : int | None
        The order in which the file-group appears in the row it belongs to
        (starting at 0). Typically corresponds to the acquisition order for
        scans within an imaging session. Can be used to distinguish between
        scans with the same series description (e.g. multiple BOLD or T1w
        scans) in the same imaging sessions.
    quality : str
        The quality label assigned to the file_group (e.g. as is saved on XNAT)
    row : DataRow
        The data row within a dataset that the file-group belongs to
    exists : bool
        Whether the file_group exists or is just a placeholder for a sink
    provenance : Provenance | None
        The provenance for the pipeline that generated the file-group,
        if applicable
    """

    path: str = attrs.field()
    uri: str = attrs.field(default=None)
    order: int = attrs.field(default=None)
    quality: DataQuality = attrs.field(default=DataQuality.usable)
    exists: bool = attrs.field(default=True)
    provenance: ty.Dict[str, ty.Any] = attrs.field(default=None)
    row: DataRow = attrs.field(default=None)

    SUBPACKAGE = "data"

    @abstractmethod
    def get(self, assume_exists=False):
        """Pulls data from the store (if remote) and caches locally

        Parameters
        ----------
        assume_exists: bool
            If set, checks to see whether the item exists are skipped (used
            to pull data after a successful workflow run)
        """
        raise NotImplementedError

    @abstractmethod
    def put(self, value):
        """Updates the value of the item in the store to the provided value,
        pushing remotely if necessary.

        Parameters
        ----------
        value : ty.Any
            The value to update
        """
        raise NotImplementedError

    @classproperty
    def desc(cls):
        """Descriptive name that can be overridden in subclasses

        Returns
        -------
        _type_
            _description_
        """
        return cls.__name__.lower()

    @property
    def recorded_checksums(self):
        if self.provenance is None:
            return None
        else:
            return self.provenance.outputs[self.name_path]

    @provenance.validator
    def check_provenance(self, _, provenance):
        "Checks that the data item path is present in the provenance"
        if provenance is not None:
            if self.path not in provenance.outputs:
                raise ArcanaNameError(
                    self.path,
                    f"{self.path} was not found in outputs "
                    f"{provenance.outputs.keys()} of provenance provenance "
                    f"{provenance}",
                )

    def _check_exists(self):
        if not self.exists:
            raise ArcanaDataNotDerivedYetError(
                self.path, f"Cannot access {self} as it hasn't been derived yet"
            )

    def _check_part_of_row(self):
        if self.row is None:
            raise ArcanaUsageError(
                f"Cannot 'get' {self} as it is not part of a dataset"
            )

    @classmethod
    def class_name(cls):
        return cls.__name__.lower()


@attrs.define
class FileGroup(DataType, metaclass=ABCMeta):
    """
    A representation of a file_group within the dataset.

    Parameters
    ----------
    name_path : str
        The name_path to the relative location of the file group, i.e. excluding
        information about which row in the data tree it belongs to
    order : int | None
        The order in which the file-group appears in the row it belongs to
        (starting at 0). Typically corresponds to the acquisition order for
        scans within an imaging session. Can be used to distinguish between
        scans with the same series description (e.g. multiple BOLD or T1w
        scans) in the same imaging sessions.
    quality : str
        The quality label assigned to the file_group (e.g. as is saved on XNAT)
    row : DataRow
        The data row within a dataset that the file-group belongs to
    exists : bool
        Whether the file_group exists or is just a placeholder for a sink
    provenance : Provenance | None
        The provenance for the pipeline that generated the file-group,
        if applicable
    fs_path : str | None
        Path to the primary file or directory on the local file system
    side_cars : ty.Dict[str, str] | None
        Additional files in the file_group. Keys should match corresponding
        side_cars dictionary in format.
    checksums : ty.Dict[str, str]
        A checksums of all files within the file_group in a dictionary sorted
        bys relative file name_paths
    """

    ext = None  # to be overridden by file-format specific classes

    fs_path: str = attrs.field(default=None, converter=optional(absolute_path))
    _checksums: ty.Dict[str, str] = attrs.field(default=None, repr=False, init=False)
    # Alternative names for the file format, empty by default overridden in
    # sub-classes where necessary
    alternative_names = ()

    @fs_path.validator
    def validate_fs_path(self, _, fs_path):
        if fs_path is not None:
            if not fs_path.exists:
                raise ArcanaUsageError(
                    "Attempting to set a path that doesn't exist " f"({fs_path})"
                )
            if not self.exists:
                raise ArcanaUsageError(
                    "Attempting to set a path to a file group that hasn't "
                    f"been derived yet ({fs_path})"
                )

    def get(self, assume_exists=False):
        if assume_exists:
            self.exists = True
        self._check_part_of_row()
        fs_paths = self.row.dataset.store.get_file_group_paths(self)
        self.exists = True
        self.set_fs_paths(fs_paths)
        self.validate_file_paths()

    def put(self, *fs_paths):
        self._check_part_of_row()
        fs_paths = [Path(p) for p in fs_paths]
        dir_paths = list(p for p in fs_paths if p.is_dir())
        if len(dir_paths) > 1:
            dir_paths_str = "', '".join(str(p) for p in dir_paths)
            raise ArcanaFileFormatError(
                f"Cannot put more than one directory, {dir_paths_str}, as part "
                f"of the same file group {self}"
            )
        # Make a copy of the file-group to validate the local paths and auto-gen
        # any defaults before they are pushed to the store
        cpy = copy(self)
        cpy.exists = True
        cpy.set_fs_paths(fs_paths)
        cache_paths = self.row.dataset.store.put_file_group_paths(self, cpy.fs_paths)
        # Set the paths to the cached files
        self.exists = True
        self.set_fs_paths(cache_paths)
        self.validate_file_paths()
        # Save provenance
        if self.provenance:
            self.row.dataset.store.put_provenance(self)

    @property
    def fs_paths(self):
        """All base paths (i.e. not nested within directories) in the file group"""
        if self.fs_path is None:
            raise ArcanaUsageError(
                f"Attempting to access file path of {self} before it is set"
            )
        return [self.fs_path]

    @classmethod
    def fs_names(cls):
        """Return names for each top-level file-system path in the file group,
        used when generating Pydra task interfaces.

        Returns
        -------
        tuple[str]
            sequence of names for top-level file-system paths in the file group"""
        return ("fs_path",)

    @classmethod
    def matches_format_name(cls, name: str):
        """Checks to see whether the provided name is a valid name for the
        file format. Alternative names can be provided for format-specific
        subclasses, or this method can be overridden. Matches are case
        insensitive.

        Parameters
        ----------
        name : str
            Name to match

        Returns
        -------
        bool
            whether or not the name matches the datatype
        """
        return name.lower() in [
            n.lower() for n in (cls.class_name(),) + cls.alternative_names
        ]

    @property
    def value(self):
        return str(self.fs_path)

    @property
    def checksums(self):
        if self._checksums is None:
            self.get_checksums()
        return self._checksums

    def get_checksums(self, force_calculate=False):
        self._check_exists()
        # Load checksums from store (e.g. via API)
        if self.row is not None and not force_calculate:
            self._checksums = self.row.dataset.store.get_checksums(self)
        # If the store cannot calculate the checksums do them manually
        else:
            self._checksums = self.calculate_checksums()

    def calculate_checksums(self):
        self._check_exists()
        checksums = {}
        for fpath in self.all_file_paths():
            fhash = hashlib.md5()
            with open(fpath, "rb") as f:
                # Calculate hash in chunks so we don't run out of memory for
                # large files.
                for chunk in iter(lambda: f.read(HASH_CHUNK_SIZE), b""):
                    fhash.update(chunk)
            checksums[fpath] = fhash.hexdigest()
        checksums = self.generalise_checksum_keys(checksums)
        return checksums

    def contents_equal(self, other, **kwargs):
        """
        Test the equality of the file_group contents with another file_group.
        If the file_group's format implements a 'contents_equal' method than
        that is used to determine the equality, otherwise a straight comparison
        of the checksums is used.

        Parameters
        ----------
        other : FileGroup
            The other file_group to compare to
        """
        self._check_exists()
        other._check_exists()
        return self.checksums[self.fs_path.name] == other.checksums[other.fs_path.name]

    @classmethod
    def resolve(cls, unresolved):
        """Resolve file group loaded from a repository to the specific datatype

        Parameters
        ----------
        unresolved : UnresolvedFileGroup
            A file group loaded from a repository that has not been resolved to
            a specific datatype yet

        Returns
        -------
        FileGroup
            The resolved file-group object

        Raises
        ------
        ArcanaUnresolvableFormatException
            If there doesn't exist a unique resolution from the unresolved file
            group to the given datatype, then an ArcanaFileFormatError should be
            raised
        """
        # Perform matching based on resource names in multi-datatype
        # file-group
        if unresolved.uris is not None:
            item = None
            for format_name, uri in unresolved.uris.items():
                if cls.matches_format_name(format_name):
                    item = cls(uri=uri, **unresolved.item_kwargs)
            if item is None:
                raise ArcanaFileFormatError(
                    f"Could not file a matching resource in {unresolved.path} for"
                    f" the given datatype ({cls.class_name()}), found "
                    "('{}')".format("', '".join(unresolved.uris))
                )
        else:
            item = cls(**unresolved.item_kwargs)
            item.set_fs_paths(unresolved.file_paths)
        return item

    @abstractmethod
    def set_fs_paths(self, fs_paths: ty.List[Path]):
        """Set the file paths of the file group

        Parameters
        ----------
        fs_paths : list[Path]
            The candidate paths from which to set the paths of the
            file group from. Note that not all paths need to be set if
            they are not relevant.

        Raises
        ------
        ArcanaFileFormatError
            is raised if the required the paths cannot be set from the provided
        """

    @classmethod
    def from_fs_paths(cls, *fs_paths: ty.List[Path], path=None):
        """Create a FileGroup object from a set of file-system paths

        Parameters
        ----------
        fs_paths : list[Path]
            The candidate paths from which to set the paths of the
            file group from. Note that not all paths need to be set if
            they are not relevant.
        path : str, optional
            the location of the file-group relative to the node it (will)
            belong to. Defaults to

        Returns
        -------
        FileGroup
            The created file-group
        """
        if path is None:
            path = fs_paths[0].stem
        obj = cls(path)
        obj.set_fs_paths(fs_paths)
        return obj

    @classmethod
    def matches_ext(cls, *paths, ext=None):
        """Returns the path out of the candidates provided that matches the
        given extension (by default the extension of the class)

        Parameters
        ----------
        *paths: list[Path]
            The paths to select from
        ext: str or None
            the extension to match (defaults to 'ext' attribute of class)

        Returns
        -------
        Path
            the matching path

        Raises
        ------
        ArcanaFileFormatError
            When no paths match or more than one path matches the given extension"""
        if ext is None:
            ext = cls.ext
        if ext:
            matches = [str(p) for p in paths if str(p).endswith("." + ext)]
        else:
            matches = paths
        if not matches:
            paths_str = ", ".join(str(p) for p in paths)
            raise ArcanaFileFormatError(
                f"No matching files with '{ext}' extension found in "
                f"file group {paths_str}"
            )
        elif len(matches) > 1:
            matches_str = ", ".join(str(p) for p in matches)
            raise ArcanaFileFormatError(
                f"Multiple files with '{ext}' extension found in : {matches_str}"
            )
        return str(matches[0])

    def validate_file_paths(self):
        attrs.validate(self)
        self.exists = True

    def _check_paths_exist(self, fs_paths: ty.List[Path]):
        if missing := [p for p in fs_paths if not p or not Path(p).exists()]:
            missing_str = "\n".join(str(p) for p in missing)
            all_str = "\n".join(str(p) for p in fs_paths)
            msg = (
                f"The following file system paths provided to {self} do not "
                f"exist:\n{missing_str}\n\nFrom full list:\n{all_str}"
            )
            for fs_path in missing:
                if fs_path:
                    if fs_path.parent.exists():
                        msg += (
                            f"\n\nFiles in the directory '{str(fs_path.parent)}' are:\n"
                        )
                        msg += "\n".join(str(p) for p in fs_path.parent.iterdir())
            raise ArcanaFileFormatError(msg)

    def convert_to(self, to_format, **kwargs):
        """Convert the FileGroup to a new datatype

        Parameters
        ----------
        to_format : type
            the file-group datatype to convert to
        **kwargs
            args to pass to the conversion process

        Returns
        -------
        FileGroup
            the converted file-group
        """
        task = to_format.converter_task(
            from_format=type(self), name="converter", **kwargs
        )
        task.inputs.to_convert = self
        result = task(plugin="serial")
        return result.output.converted

    @classmethod
    def converter_task(cls, from_format, name, **kwargs):
        """Adds a converter row to a workflow

        Parameters
        ----------
        from_format : type
            the file-group class to convert from
        taks_name: str
            the name for the converter task
        **kwargs: dict[str, ty.Any]
            keyword arguments passed through to the converter

        Returns
        -------
        Workflow
            Pydra workflow to perform the conversion with an input called
            'to_convert' and an output called 'converted', which take and
            produce file-groups in `from_format` and `cls` types respectively.
        """

        wf = Workflow(name=name, input_spec=["to_convert"])

        # Get row to extract paths from file-group lazy field
        wf.add(
            func_task(
                access_paths,
                in_fields=[("from_format", type), ("file_group", from_format)],
                out_fields=[(i, Path) for i in from_format.fs_names()],
                # name='extract',
                from_format=from_format,
                file_group=wf.lzin.to_convert,
            )
        )

        # Aggregate converter inputs and combine with fixed keyword args
        conv_inputs = {
            n: getattr(wf.access_paths.lzout, n) for n in from_format.fs_names()
        }
        conv_inputs.update(kwargs)
        # Create converter node
        converter, output_lfs = cls.find_converter(from_format)(**conv_inputs)
        # If there is only one output lazy field, place it in a tuple so it can
        # be zipped with cls.fs_names()
        if isinstance(output_lfs, LazyField):
            output_lfs = (output_lfs,)
        # converter.name = 'converter'
        # for lf in output_lfs:
        #     lf.name = 'converter'
        wf.add(converter)

        # Encapsulate output paths from converter back into a file group object
        to_encapsulate = dict(zip(cls.fs_names(), output_lfs))

        logger.debug("Paths to encapsulate are:\n%s", to_encapsulate)

        wf.add(
            func_task(
                encapsulate_paths,
                in_fields=[("to_format", type), ("to_convert", from_format)]
                + [(o, ty.Union[str, Path]) for o in cls.fs_names()],
                out_fields=[("converted", cls)],
                # name='encapsulate',
                to_format=cls,
                to_convert=wf.lzin.to_convert,
                **to_encapsulate,
            )
        )

        wf.set_output(("converted", wf.encapsulate_paths.lzout.converted))

        return wf

    @classmethod
    def find_converter(cls, from_format):
        """Selects the converter method from the given datatype. Will select the
        most specific conversion.

        Parameters
        ----------
        from_format : type
            The datatype type to convert from

        Returns
        -------
        function or NoneType
            The bound method that adds rows to a given workflow if conversion is required
            and None if no conversion is required

        Raises
        ------
        ArcanaFormatConversionError
            _description_
        """
        if from_format is cls or issubclass(from_format, cls):
            return None  # No conversion is required
        converter = None
        for attr_name in dir(cls):
            meth = getattr(cls, attr_name)
            try:
                converts_from = meth.__annotations__[CONVERTER_ANNOTATIONS]
            except (AttributeError, KeyError):
                pass
            else:
                if from_format is converts_from or issubclass(
                    from_format, converts_from
                ):
                    if converter:
                        prev_converts_from = converter.__annotations__[
                            CONVERTER_ANNOTATIONS
                        ]
                        if issubclass(converts_from, prev_converts_from):
                            converter = meth
                        elif not issubclass(prev_converts_from, converts_from):
                            raise ArcanaFormatConversionError(
                                f"Ambiguous converters between {from_format} "
                                f"and {cls}: {converter} and {meth}. Please "
                                f"define a specific converter from {from_format} "
                                f"(i.e. instead of from {prev_converts_from} "
                                f"and {converts_from} respectively)"
                            )
                    else:
                        converter = meth
        if converter is None:
            raise ArcanaFormatConversionError(
                f"No datatype converters are defined from {from_format} to {cls}"
            )
        return converter

    def generalise_checksum_keys(
        self, checksums: ty.Dict[str, str], base_path: Path = None
    ):
        """Generalises the paths used for the file paths in a checksum dictionary
        so that they are the same irrespective of that the top-level file-system
        paths are

        Parameters
        ----------
        checksums: dict[str, str]
            The checksum dict mapping relative file paths to checksums

        Returns
        -------
        dict[str, str]
            The checksum dict with file paths generalised"""
        if base_path is None:
            base_path = self.fs_path
        return {str(Path(k).relative_to(base_path)): v for k, v in checksums.items()}

    @classmethod
    def access_contents_task(cls, file_group_lf: LazyField):
        """Access the fs paths of the file group"""

    @classmethod
    def from_fs_path(cls, fs_path):
        file_group = cls(path=Path(fs_path).stem)
        file_group.set_fs_paths([fs_path])
        return file_group

    @classmethod
    def append_ext(cls, path: Path):
        if path.ext is not None:
            path = path.with_suffix(cls.ext)
        return path

    @classmethod
    def all_exts(cls):
        return [""]


def access_paths(from_format, file_group):
    """Copies files into the CWD renaming so the basenames match
    except for extensions"""
    logger.debug(
        "Extracting paths from %s (%s format) before conversion",
        file_group,
        from_format,
    )
    cpy = file_group.copy_to(path2varname(file_group.path), symlink=True)
    return cpy.fs_paths if len(cpy.fs_paths) > 1 else cpy.fs_path


def encapsulate_paths(
    to_format: type, to_convert: FileGroup, **fs_paths: ty.List[Path]
):
    """Copies files into the CWD renaming so the basenames match
    except for extensions"""
    logger.debug(
        "Encapsulating %s into %s format after conversion", fs_paths, to_format
    )
    file_group = to_format(to_convert.path + "_" + to_format.class_name())
    file_group.set_fs_paths(fs_paths.values())
    return file_group


@attrs.define
class Field(DataType):
    """
    A representation of a value field in the dataset.

    Parameters
    ----------
    name_path : str
        The name_path to the relative location of the field, i.e. excluding
        information about which row in the data tree it belongs to
    derived : bool
        Whether or not the value belongs in the derived session or not
    row : DataRow
        The data row that the field belongs to
    exists : bool
        Whether the field exists or is just a placeholder for a sink
    provenance : Provenance | None
        The provenance for the pipeline that generated the field,
        if applicable
    """

    value: ty.Any = None

    def get(self, assume_exists=False):
        if not assume_exists:
            self._check_exists()
        self._check_part_of_row()
        self.value = self.row.dataset.store.get_field_value(self)

    def put(self, value):
        self._check_part_of_row()
        self.row.dataset.store.put_field_value(self, self.format(value))
        self.exists = True

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __str__(self):
        if self.datatype.__args__:  # Sequence type
            val = "[" + ",".join(self._to_str(v) for v in self.value) + "]"
        else:
            val = self._to_str(self.value)
        return val

    def _to_str(self, val):
        if self.datatype is str:
            val = '"{}"'.format(val)
        else:
            val = str(val)
        return val

    def get_checksums(self):
        """
        For duck-typing with file_groups in checksum management. Instead of a
        checksum, just the value of the field is used
        """
        return self.value



from __future__ import annotations
import logging
from abc import abstractmethod, ABCMeta
from pathlib import Path
import attrs
import typing as ty
import yaml
from arcana.core.utils.serialize import (
    asdict,
    fromdict,
)
import arcana
from arcana.core.utils.misc import get_config_file_path
from arcana.core.utils.packaging import list_subclasses
from arcana.core.exceptions import ArcanaUsageError, ArcanaNameError

DS = ty.TypeVar("DS", bound="DataStore")

logger = logging.getLogger("arcana")


@attrs.define
class DataStore(metaclass=ABCMeta):
    # """
    # Abstract base class for all Repository systems, DaRIS, XNAT and
    # local file system. Sets out the interface that all Repository
    # classes should implement.
    # """

    _connection_depth = attrs.field(
        default=0, init=False, hash=False, repr=False, eq=False
    )

    CONFIG_NAME = "stores"

    SUBPACKAGE = "data"

    @abstractmethod
    def find_rows(self, dataset):
        """
        Find all data rows for a dataset in the store and populate the
        Dataset object using its `add_row` method.

        Parameters
        ----------
        dataset : Dataset
            The dataset to populate with rows
        """

    @abstractmethod
    def find_items(self, row):
        """
        Find all data items within a data row and populate the DataRow object
        with them using the `add_file_group` and `add_field` methods.

        Parameters
        ----------
        row : DataRow
            The data row to populate with items
        """

    @abstractmethod
    def get_file_group_paths(self, file_group, cache_only=False):
        """
        Cache the file_group locally (if required) and return the locations
        of the cached primary file and side cars

        Parameters
        ----------
        file_group : FileGroup
            The file_group to cache locally
        cache_only : bool
            Whether to attempt to extract the file groups from the local cache
            (if applicable) and raise an error otherwise

        Returns
        -------
        fs_paths : list[str]
            The file-system path to the cached files

        Raises
        ------
        ArcanaCacheError
            If cache_only is set and there is a mismatch between the cached
            and remote versions
        """

    @abstractmethod
    def get_field_value(self, field):
        """
        Extract and return the value of the field from the store

        Parameters
        ----------
        field : Field
            The field to retrieve the value for

        Returns
        -------
        value : int | float | str | ty.List[int] | ty.List[float] | ty.List[str]
            The value of the Field
        """

    @abstractmethod
    def put_file_group_paths(self, file_group, fs_paths):
        """
        Inserts or updates the file_group into the store

        Parameters
        ----------
        file_group : FileGroup
            The file_group to insert into the store
        fs_paths : list[Path]
            The file-system paths to the files/directories to sync

        Returns
        -------
        cached_paths : list[str]
            The paths of the files where they are cached in the file system
        """

    @abstractmethod
    def put_field_value(self, field, value):
        """
        Inserts or updates the fields into the store

        Parameters
        ----------
        field : Field
            The field to insert into the store
        """

    @abstractmethod
    def save_dataset_definition(
        self, dataset_id: str, definition: ty.Dict[str, ty.Any], name: str
    ):
        """Save definition of dataset within the store

        Parameters
        ----------
        dataset_id: str
            The ID/path of the dataset within the store
        definition: dict[str, Any]
            A dictionary containing the dct Dataset to be saved. The
            dictionary is in a format ready to be dumped to file as JSON or
            YAML.
        name: str
            Name for the dataset definition to distinguish it from other
            definitions for the same directory/project"""

    @abstractmethod
    def load_dataset_definition(
        self, dataset_id: str, name: str
    ) -> ty.Dict[str, ty.Any]:
        """Load definition of a dataset saved within the store

        Parameters
        ----------
        dataset_id: str
            The ID (e.g. file-system path, XNAT project ID) of the project
        name: str
            Name for the dataset definition to distinguish it from other
            definitions for the same directory/project

        Returns
        -------
        definition: dict[str, Any]
            A dct Dataset object that was saved in the data store
        """

    @abstractmethod
    def put_provenance(self, item, provenance: ty.Dict[str, ty.Any]):
        """Stores provenance information for a given data item in the store

        Parameters
        ----------
        item: DataType
            The item to store the provenance data for
        provenance: dict[str, Any]
            The provenance data to store"""

    @abstractmethod
    def get_provenance(self, item) -> ty.Dict[str, ty.Any]:
        """Stores provenance information for a given data item in the store

        Parameters
        ----------
        item: DataType
            The item to store the provenance data for

        Returns
        -------
        provenance: dict[str, Any] or None
            The provenance data stored in the repository for the data item.
            None if no provenance data has been stored"""

    def get_checksums(self, file_group):
        """
        Override this method to return checksums for files that are stored
        with remote files (e.g. in XNAT). If no checksums are stored in the
        store then just leave this method to just access the file and
        recalculate them.

        Parameters
        ----------
        file_group : FileGroup
            The file_group to return the checksums for

        Returns
        -------
        checksums : dct[str, str]
            A dictionary with keys corresponding to the relative paths of all
            files in the file_group from the base path and values equal to the
            MD5 hex digest. The primary file in the file-set (i.e. the one that
            the path points to) should be specified by '.'.
        """
        return file_group.calculate_checksums()

    def connect(self):
        """
        If a connection session is required to the store manage it here
        """

    def disconnect(self):
        """
        If a connection session is required to the store manage it here
        """

    def save(self, name: str, config_path: Path = None):
        """Saves the configuration of a DataStore in 'stores.yaml'

        Parameters
        ----------
        name
            The name under which to save the data store
        config_path : Path, optional
            the path to save the config file to, defaults to `~/.arcana/stores.yaml`
        """
        if name in self.singletons():
            raise ArcanaNameError(
                name, f"Name '{name}' clashes with built-in type of store"
            )
        entries = self.load_saved_entries()
        # connect to store in case it is needed in the asdict method and to
        # test the connection in general before it is saved
        with self:
            entries[name] = self.asdict()
        self.save_entries(entries, config_path=config_path)

    def asdict(self, **kwargs):
        return asdict(self, **kwargs)

    def site_licenses_dataset(self):
        """Can be overridden by subclasses to provide a dataset to hold site-wide licenses"""
        return None

    @classmethod
    def load(cls: ty.Type[DS], name: str, config_path: Path = None, **kwargs) -> DS:
        """Loads a DataStore from that has been saved in the configuration file.
        If no entry is saved under that name, then it searches for DataStore
        sub-classes with aliases matching `name` and checks whether they can
        be initialised without any parameters.

        Parameters
        ----------
        name : str
            Name that the store was saved under
        config_path : Path, optional
            path to the config file, defaults to `~/.arcana/stores.yaml`
        **kwargs
            keyword args passed to the store, overriding values stored in the
            entry

        Returns
        -------
        DataStore
            The data store retrieved from the stores.yaml file

        Raises
        ------
        ArcanaNameError
            If the name is not found in the saved stores
        """
        entries = cls.load_saved_entries(config_path)
        try:
            entry = entries[name]
        except KeyError:
            try:
                return cls.singletons()[name]
            except KeyError:
                raise ArcanaNameError(
                    name, f"No saved data store or built-in type matches '{name}'"
                )
        else:
            entry.update(kwargs)
            store = fromdict(entry)  # Would be good to use a class resolver here
        return store

    @classmethod
    def remove(cls, name: str, config_path: Path = None):
        """Removes the entry saved under 'name' in the config file

        Parameters
        ----------
        name
            Name of the configuration to remove
        """
        entries = cls.load_saved_entries(config_path)
        del entries[name]
        cls.save_entries(entries)

    def new_dataset(self, id, hierarchy=None, space=None, **kwargs):
        """
        Returns a dataset from the XNAT repository

        Parameters
        ----------
        id : str
            The ID (or file-system path) of the project (or directory) within
            the store
        space: DataSpace
            The data space of the dataset
        hierarchy: list[DataSpace or str]
            The hierarchy of the dataset
        space : EnumMeta
            The DataSpace enum that defines the frequencies (e.g.
            per-session, per-subject,...) present in the dataset.
        **kwargs:
            Keyword args passed on to the Dataset init method
        """
        from .space import DataSpace

        if not hierarchy:
            if space:
                hierarchy = [max(space)]
            else:
                try:
                    hierarchy = self.DEFAULT_HIERARCHY
                except AttributeError as e:
                    raise ArcanaUsageError(
                        "'hierarchy' kwarg must be specified for datasets in "
                        f"{type(self)} stores"
                    ) from e
        if not space:
            if hierarchy and isinstance(hierarchy[0], DataSpace):
                space = type(hierarchy[0])
            else:
                try:
                    space = self.DEFAULT_SPACE
                except AttributeError as e:
                    raise ArcanaUsageError(
                        "'space' kwarg must be specified for datasets in "
                        f"{type(self)} stores"
                    ) from e
        from arcana.core.data.set import (
            Dataset,
        )  # avoid circular imports it is imported here rather than at the top of the file

        dataset = Dataset(id, store=self, space=space, hierarchy=hierarchy, **kwargs)
        return dataset

    def load_dataset(self, id, name=None):
        from arcana.core.data.set import (
            Dataset,
        )  # avoid circular imports it is imported here rather than at the top of the file

        if name is None:
            name = Dataset.DEFAULT_NAME
        dct = self.load_dataset_definition(id, name)
        if dct is None:
            raise KeyError(f"Did not find a dataset '{id}::{name}'")
        return fromdict(dct, id=id, name=name, store=self)

    @classmethod
    def singletons(cls):
        """Returns stores in a dictionary indexed by their aliases, for which there
        only needs to be a single instance"""
        try:
            return cls._singletons
        except AttributeError:
            pass
        # If not saved in the configuration file search for sub-classes
        # whose alias matches `name` and can be initialised without params
        cls._singletons = {}
        for store_cls in list_subclasses(arcana, DataStore, subpkg="data"):
            try:
                cls._singletons[store_cls.get_alias()] = store_cls()
            except Exception:
                pass
        return cls._singletons

    @classmethod
    def load_saved_entries(cls, config_path: Path = None):
        if config_path is None:
            config_path = get_config_file_path(cls.CONFIG_NAME)
        if config_path.exists():
            with open(config_path) as f:
                entries = yaml.load(f, Loader=yaml.Loader)
        else:
            entries = {}
        return entries

    @classmethod
    def save_entries(cls, entries, config_path: Path = None):
        if config_path is None:
            config_path = get_config_file_path(cls.CONFIG_NAME)
        with open(config_path, "w") as f:
            yaml.dump(entries, f)

    def __enter__(self):
        # This allows the store to be used within nested contexts
        # but still only use one connection. This is useful for calling
        # methods that need connections, and therefore control their
        # own connection, in batches using the same connection by
        # placing the batch calls within an outer context.
        if self._connection_depth == 0:
            self.connect()
        self._connection_depth += 1
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self._connection_depth -= 1
        if self._connection_depth == 0:
            self.disconnect()

    @classmethod
    def get_alias(cls):
        try:
            alias = cls.alias
        except AttributeError:
            alias = cls.__name__.lower()
        return alias



import pytest
import os.path
from unittest.mock import patch
from arcana.core.data.set import Dataset
from arcana.core.analysis.salience import DataQuality, ColumnSalience
from arcana.core.utils.testing.data import TestDataSpace
from arcana.core.cli.dataset import define, add_source, add_sink, missing_items
from arcana.dirtree.data import Text
from arcana.core.utils.testing import make_dataset_locator, show_cli_trace


ARBITRARY_INTS_A = [234221, 93380, 43271, 137483, 30009, 214205, 363526]
ARBITRARY_INTS_B = [353726, 29202, 32867, 129872, 12281, 776524, 908763]


def get_arbitrary_slice(i, dim_length):
    a = ARBITRARY_INTS_A[i] % dim_length
    b = ARBITRARY_INTS_B[i] % dim_length
    lower = min(a, b)
    upper = max(a, b) + 1
    return lower, upper


def test_add_column_cli(saved_dataset, cli_runner):
    dataset_locator = make_dataset_locator(saved_dataset)
    # Get CLI name for dataset (i.e. file system path prepended by 'file//')
    # Add source to loaded dataset
    saved_dataset.add_source(
        name="a_source",
        path="file1",
        datatype=Text,
        row_frequency=TestDataSpace.d,
        quality_threshold=DataQuality.questionable,
        order=1,
        header_vals={},
        is_regex=False,
    )
    # Add source column to saved dataset
    result = cli_runner(
        add_source,
        [
            dataset_locator,
            "a_source",
            "common:Text",
            "--path",
            "file1",
            "--row-frequency",
            "d",
            "--quality",
            "questionable",
            "--order",
            "1",
            "--no-regex",
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    # Add source to loaded dataset
    saved_dataset.add_sink(
        name="a_sink",
        path="deriv",
        datatype=Text,
        row_frequency=TestDataSpace.d,
        salience=ColumnSalience.qa,
    )
    result = cli_runner(
        add_sink,
        [
            dataset_locator,
            "a_sink",
            "common:Text",
            "--path",
            "deriv",
            "--row-frequency",
            "d",
            "--salience",
            "qa",
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    # Reload the saved dataset and check the parameters were saved/loaded
    # correctly
    loaded_dataset = Dataset.load(dataset_locator)
    assert saved_dataset.columns == loaded_dataset.columns


@pytest.mark.skip("Not implemented")
def test_add_missing_items_cli(dataset, cli_runner):
    result = cli_runner(missing_items, [])
    assert result.exit_code == 0, show_cli_trace(result)


def test_define_cli(dataset, cli_runner):
    blueprint = dataset.__annotations__["blueprint"]
    # Get CLI name for dataset (i.e. file system path prepended by 'file//')
    path = "file//" + os.path.abspath(dataset.id)
    # Start generating the arguments for the CLI
    args = [str(h) for h in blueprint.hierarchy]
    # Generate "arbitrary" values for included and excluded from dim length
    # and index
    included = []
    excluded = []
    for i, (dim_length, axis) in enumerate(zip(blueprint.dim_lengths, dataset.space)):
        a, b = get_arbitrary_slice(i, dim_length)
        if i % 2:
            included.append((axis, f"{a}:{b}"))
        elif (b - a) < dim_length:  # Check that we aren't excluding all
            excluded.append((axis, f"{a}:{b}"))
    # Add include and exclude options
    for axis, slce in included:
        args.extend(["--include", str(axis), slce])
    for axis, slce in excluded:
        args.extend(["--exclude", str(axis), slce])
    args.extend(["--space", "arcana.core.utils.testing.data:TestDataSpace"])
    # Run the command line
    result = cli_runner(define, [path, *args])
    # Check tool completed successfully
    assert result.exit_code == 0, show_cli_trace(result)
    # Reload the saved dataset and check the parameters were saved/loaded
    # correctly
    loaded_dataset = Dataset.load(path)
    assert loaded_dataset.hierarchy == blueprint.hierarchy
    assert loaded_dataset.include == included
    assert loaded_dataset.exclude == excluded



from __future__ import annotations
import os
import shutil
import typing as ty
from itertools import product
import zipfile
from pathlib import Path
import attrs
from dataclasses import dataclass, field as dataclass_field
from pydra import mark
from arcana.core.utils.misc import set_cwd, path2varname
from arcana.core.data.space import DataSpace
from arcana.core.data.store import DataStore
from arcana.core.data.type.file import WithSideCars, BaseFile
from arcana.dirtree.data import FileSystem
from arcana.dirtree.data.formats import Text
from arcana.core.mark import converter


class TestDataSpace(DataSpace):
    """Dummy data dimensions for ease of testing"""

    # Per dataset
    _ = 0b0000

    # Basis
    a = 0b1000
    b = 0b0100
    c = 0b0010
    d = 0b0001

    # Secondary combinations
    ab = 0b1100
    ac = 0b1010
    ad = 0b1001
    bc = 0b0110
    bd = 0b0101
    cd = 0b0011

    # Tertiary combinations
    abc = 0b1110
    abd = 0b1101
    acd = 0b1011
    bcd = 0b0111

    # Leaf rows
    abcd = 0b1111


class Xyz(WithSideCars):

    ext = "x"
    side_car_exts = ("y", "z")


# -----------------------
# Test dataset structures
# -----------------------


@dataclass
class TestDatasetBlueprint:

    hierarchy: ty.List[DataSpace]
    dim_lengths: ty.List[int]  # size of layers a-d respectively
    files: ty.List[str]  # files present at bottom layer
    id_inference: ty.List[ty.Tuple[DataSpace, str]] = dataclass_field(
        default_factory=list
    )  # id_inference dict
    expected_formats: ty.Dict[str, ty.Tuple[type, ty.List[str]]] = dataclass_field(
        default_factory=dict
    )  # expected formats
    derivatives: ty.List[
        ty.Tuple[str, DataSpace, type, ty.List[str]]
    ] = dataclass_field(
        default_factory=list
    )  # files to insert as derivatives

    @property
    def space(self):
        return type(self.hierarchy[0])


def make_dataset(
    blueprint: TestDatasetBlueprint, dataset_path: Path, source_data: Path = None
):
    create_dataset_data_in_repo(blueprint, dataset_path, source_data=source_data)
    return access_dataset(blueprint, dataset_path)


def create_dataset_data_in_repo(
    blueprint: TestDatasetBlueprint, dataset_path: Path, source_data: Path = None
):
    "Creates a dataset from parameters in TEST_DATASETS"
    dataset_path.mkdir(exist_ok=True, parents=True)
    for id_tple in product(*(list(range(d)) for d in blueprint.dim_lengths)):
        ids = dict(zip(blueprint.space.axes(), id_tple))
        dpath = dataset_path
        for layer in blueprint.hierarchy:
            dpath /= "".join(f"{b}{ids[b]}" for b in layer.span())
        os.makedirs(dpath)
        for fname in blueprint.files:
            if source_data is not None:
                src_path = source_data.joinpath(*fname.split("/"))
                parts = fname.split(".")
                dst_path = dpath / (path2varname(parts[0]) + "." + ".".join(parts[1:]))
                dst_path.parent.mkdir(exist_ok=True)
                if src_path.is_dir():
                    shutil.copytree(src_path, dst_path)
                else:
                    shutil.copyfile(src_path, dst_path, follow_symlinks=True)
            else:
                create_test_file(fname, dpath)


def access_dataset(blueprint, dataset_path):
    dataset = FileSystem().new_dataset(
        dataset_path,
        hierarchy=blueprint.hierarchy,
        id_inference=blueprint.id_inference,
    )
    dataset.__annotations__["blueprint"] = blueprint
    return dataset


def create_test_file(fname, dpath):
    dpath = Path(dpath)
    os.makedirs(dpath, exist_ok=True)
    next_part = fname
    if next_part.endswith(".zip"):
        next_part = next_part.strip(".zip")
    fpath = Path(next_part)
    # Make double dir
    if next_part.startswith("doubledir"):
        os.makedirs(dpath / fpath, exist_ok=True)
        next_part = "dir"
        fpath /= next_part
    if next_part.startswith("dir"):
        os.makedirs(dpath / fpath, exist_ok=True)
        next_part = "test.txt"
        fpath /= next_part
    if not fpath.suffix:
        fpath = fpath.with_suffix(".txt")
    with open(dpath / fpath, "w") as f:
        f.write(f"{fname}")
    if fname.endswith(".zip"):
        with zipfile.ZipFile(dpath / fname, mode="w") as zfile, set_cwd(dpath):
            zfile.write(fpath)
        (dpath / fpath).unlink()
        fpath = Path(fname)
    return fpath


def save_dataset(work_dir, name=None):
    blueprint = TestDatasetBlueprint(
        [
            TestDataSpace.abcd
        ],  # e.g. XNAT where session ID is unique in project but final layer is organised by timepoint
        [1, 1, 1, 1],
        ["file1.txt", "file2.txt"],
        {},
        {},
        [],
    )

    dataset_path = work_dir / "saved_dataset"
    dataset = make_dataset(blueprint, dataset_path)
    dataset.save(name)
    return dataset


class Nifti(BaseFile):

    ext = "nii"


class NiftiGz(Nifti):

    ext = "nii.gz"


class NiftiX(WithSideCars, Nifti):

    side_car_exts = ("json",)


class MrtrixImage(BaseFile):

    ext = "mif"


class Analyze(WithSideCars, BaseFile):

    ext = "img"
    side_car_exts = ("hdr",)


class NiftiGzX(NiftiX, NiftiGz):

    pass


class EncodedText(BaseFile):
    """A text file where the characters ASCII codes are shifted on conversion
    from text
    """

    ext = "enc"

    @classmethod
    @converter(Text)
    def encode(cls, fs_path: ty.Union[str, Path], shift: int = 0):
        shift = int(shift)
        node = encoder_task(in_file=fs_path, shift=shift)
        return node, node.lzout.out


class DecodedText(Text):
    @classmethod
    @converter(EncodedText)
    def decode(cls, fs_path: Path, shift: int = 0):
        shift = int(shift)
        node = encoder_task(
            in_file=fs_path, shift=-shift, out_file="out_file.txt"
        )  # Just shift it backwards by the same amount
        return node, node.lzout.out


@mark.task
def encoder_task(
    in_file: ty.Union[str, Path],
    shift: int,
    out_file: ty.Union[str, Path] = "out_file.enc",
) -> ty.Union[str, Path]:
    with open(in_file) as f:
        contents = f.read()
    encoded = encode_text(contents, shift)
    with open(out_file, "w") as f:
        f.write(encoded)
    return Path(out_file).absolute()


def encode_text(text: str, shift: int) -> str:
    encoded = []
    for c in text:
        encoded.append(chr(ord(c) + shift))
    return "".join(encoded)


@attrs.define
class MockDataStore(DataStore):
    """A mock data store to test store CLI. None of the methods do anything, not even
    return mock objects.
    """

    alias = "mock"

    server: str
    user: str
    password: str
    cache_dir: str

    def find_rows(self, dataset):
        """
        Find all data rows for a dataset in the store and populate the
        Dataset object using its `add_row` method.

        Parameters
        ----------
        dataset : Dataset
            The dataset to populate with rows
        """

    def find_items(self, row):
        """
        Find all data items within a data row and populate the DataRow object
        with them using the `add_file_group` and `add_field` methods.

        Parameters
        ----------
        row : DataRow
            The data row to populate with items
        """

    def get_file_group_paths(self, file_group, cache_only=False):
        """
        Cache the file_group locally (if required) and return the locations
        of the cached primary file and side cars

        Parameters
        ----------
        file_group : FileGroup
            The file_group to cache locally
        cache_only : bool
            Whether to attempt to extract the file groups from the local cache
            (if applicable) and raise an error otherwise

        Returns
        -------
        fs_paths : list[str]
            The file-system path to the cached files

        Raises
        ------
        ArcanaCacheError
            If cache_only is set and there is a mismatch between the cached
            and remote versions
        """

    def get_field_value(self, field):
        """
        Extract and return the value of the field from the store

        Parameters
        ----------
        field : Field
            The field to retrieve the value for

        Returns
        -------
        value : int | float | str | ty.List[int] | ty.List[float] | ty.List[str]
            The value of the Field
        """

    def put_file_group_paths(self, file_group, fs_paths):
        """
        Inserts or updates the file_group into the store

        Parameters
        ----------
        file_group : FileGroup
            The file_group to insert into the store
        fs_paths : list[Path]
            The file-system paths to the files/directories to sync

        Returns
        -------
        cached_paths : list[str]
            The paths of the files where they are cached in the file system
        """

    def put_field_value(self, field, value):
        """
        Inserts or updates the fields into the store

        Parameters
        ----------
        field : Field
            The field to insert into the store
        """

    def save_dataset_definition(
        self, dataset_id: str, definition: ty.Dict[str, ty.Any], name: str
    ):
        """Save definition of dataset within the store

        Parameters
        ----------
        dataset_id: str
            The ID/path of the dataset within the store
        definition: dict[str, Any]
            A dictionary containing the dct Dataset to be saved. The
            dictionary is in a format ready to be dumped to file as JSON or
            YAML.
        name: str
            Name for the dataset definition to distinguish it from other
            definitions for the same directory/project"""

    def load_dataset_definition(
        self, dataset_id: str, name: str
    ) -> ty.Dict[str, ty.Any]:
        """Load definition of a dataset saved within the store

        Parameters
        ----------
        dataset_id: str
            The ID (e.g. file-system path, XNAT project ID) of the project
        name: str
            Name for the dataset definition to distinguish it from other
            definitions for the same directory/project

        Returns
        -------
        definition: dict[str, Any]
            A dct Dataset object that was saved in the data store
        """

    def put_provenance(self, item, provenance: ty.Dict[str, ty.Any]):
        """Stores provenance information for a given data item in the store

        Parameters
        ----------
        item: DataType
            The item to store the provenance data for
        provenance: dict[str, Any]
            The provenance data to store"""

    def get_provenance(self, item) -> ty.Dict[str, ty.Any]:
        """Stores provenance information for a given data item in the store

        Parameters
        ----------
        item: DataType
            The item to store the provenance data for

        Returns
        -------
        provenance: dict[str, Any] or None
            The provenance data stored in the repository for the data item.
            None if no provenance data has been stored"""



from __future__ import annotations
import click
from .base import cli
from arcana.core.data.set import Dataset
from arcana.core.data.store import DataStore
from arcana.core.data.space import DataSpace
from arcana.core.data.type.base import DataType
from arcana.core.utils.serialize import ClassResolver


@cli.group()
def dataset():
    pass


@dataset.command(
    name="define",
    help=(
        """Define the tree structure and IDs to include in a
dataset. Where possible, the definition file is saved inside the dataset for
use by multiple users, if not possible it is stored in the ~/.arcana directory.

ID_STR string containing the nick-name of the store, the ID of the dataset
(e.g. XNAT project ID or file-system directory) and the dataset's name in the
format <store-nickname>//<dataset-id>[@<dataset-name>]

HIERARCHY the data frequencies that are present in the data tree. For some
store types this is fixed (e.g. XNAT-> subject > session) but for more flexible
(e.g. FileSystem), which dimension each layer of sub-directories corresponds to
can be arbitrarily specified. dimensions"""
    ),
)
@click.argument("id_str")
@click.argument("hierarchy", nargs=-1)
@click.option(
    "--space",
    type=str,
    default=None,
    help=(
        'The "space" of the dataset, defines the dimensions along the ids '
        "of row can vary"
    ),
)
@click.option(
    "--include",
    nargs=2,
    default=(),
    metavar="<freq-id>",
    multiple=True,
    type=str,
    help=(
        "The rows to include in the dataset. First value is the "
        "row-frequency of the ID (e.g. 'group', 'subject', 'session') "
        "followed by the IDs to be included in the dataset. "
        "If the second arg contains '/' then it is interpreted as "
        "the path to a text file containing a list of IDs"
    ),
)
@click.option(
    "--exclude",
    nargs=2,
    default=(),
    metavar="<freq-id>",
    multiple=True,
    type=str,
    help=(
        "The rows to exclude from the dataset. First value is the "
        "row-frequency of the ID (e.g. 'group', 'subject', 'session') "
        "followed by the IDs to be included in the dataset. "
        "If the second arg contains '/' then it is interpreted as "
        "the path to a text file containing a list of IDs"
    ),
)
@click.option(
    "--space",
    default="medimage:Clinical",
    help=(
        "The enum that specifies the data dimensions of the dataset. "
        "Defaults to `Clinical`, which "
        "consists of the typical dataset>group>subject>session "
        "data tree used in medimage trials/studies"
    ),
)
@click.option(
    "--id-inference",
    nargs=2,
    metavar="<source-regex>",
    multiple=True,
    help="""Specifies how IDs of row frequencies that not explicitly
provided are inferred from the IDs that are. For example, given a set
of subject IDs that are a combination of the ID of the group that they belong
to + their member IDs (i.e. matched test/controls have same member ID), e.g.

CONTROL01, CONTROL02, CONTROL03, ... and TEST01, TEST02, TEST03

the group ID can be extracted by providing the ID to source it from
(i.e. subject) and a regular expression (in Python regex syntax:
https://docs.python.org/3/library/re.html) with a named
groups corresponding to the inferred IDs

--id-inference subject '(?P<group>[A-Z]+)(?P<member>[0-9]+)'

""",
)
def define(id_str, hierarchy, include, exclude, space, id_inference):

    store_name, id, name = Dataset.parse_id_str(id_str)

    if not hierarchy:
        hierarchy = None

    store = DataStore.load(store_name)

    if space:
        space = ClassResolver(DataSpace)(space)

    dataset = store.new_dataset(
        id,
        hierarchy=hierarchy,
        space=space,
        id_inference=id_inference,
        include=include,
        exclude=exclude,
    )

    dataset.save(name)


@dataset.command(
    help="""
Renames a data store saved in the stores.yaml to a new name

dataset_path
    The current name of the store
new_name
    The new name for the store"""
)
@click.argument("dataset_path")
@click.argument("new_name")
def copy(dataset_path, new_name):
    dataset = Dataset.load(dataset_path)
    dataset.save(new_name)


# def optional_args(names, args):
#     kwargs = {}


@dataset.command(
    name="add-source",
    help="""Adds a source column to a dataset. A source column
selects comparable items along a dimension of the dataset to serve as
an input to pipelines and analyses.

DATASET_PATH: The path to the dataset including store and dataset name
(where applicable), e.g. central-xnat//MYXNATPROJECT:pass_t1w_qc

NAME: The name the source will be referenced by

FORMAT: The data type of the column. Can be a field (int|float|str|bool),
field array (list[int|float|str|bool]) or
"file-group" (file, file+header/side-cars or directory)
""",
)
@click.argument("dataset_path")
@click.argument("name")
@click.argument("datatype")
@click.option(
    "--row-frequency",
    "-f",
    metavar="<dimension>",
    help=(
        "The row-frequency that items appear in the dataset (e.g. per "
        "'session', 'subject', 'timepoint', 'group', 'dataset' for "
        "medimage:Clinical data dimensions"
    ),
    show_default="highest",
)
@click.option(
    "--path",
    "-p",
    help=(
        "Path to item in the dataset. If 'regex' option is provided it will "
        "be treated as a regular-expression (in Python syntax)"
    ),
)
@click.option(
    "--order",
    type=int,
    help=(
        "If multiple items match the remaining criteria within a session, "
        "select the <order>th of the matching items"
    ),
)
@click.option(
    "--quality",
    "-q",
    default="usable",
    help=(
        "For data stores that enable flagging of data quality, "
        "this option can filter out poor quality scans"
    ),
)
@click.option(
    "--regex/--no-regex",
    "is_regex",
    default=True,
    help=(
        "Whether the 'path' option should be treated as a regular expression " "or not"
    ),
)
@click.option(
    "--header",
    "-h",
    nargs=2,
    metavar="<key-val>",
    multiple=True,
    help=(
        "Match on specific header value. This option is only valid for "
        "select formats that the implement the 'header_val()' method "
        "(e.g. medimage:dicom)."
    ),
)
def add_source(
    dataset_path, name, datatype, row_frequency, path, order, quality, is_regex, header
):
    dataset = Dataset.load(dataset_path)
    dataset.add_source(
        name=name,
        path=path,
        datatype=ClassResolver(DataType)(datatype),
        row_frequency=row_frequency,
        quality_threshold=quality,
        order=order,
        header_vals=dict(header),
        is_regex=is_regex,
    )
    dataset.save()


@dataset.command(
    name="add-sink",
    help="""Adds a sink column to a dataset. A sink column
specifies how data should be written into the dataset.

Arguments
---------
dataset_path
    The path to the dataset including store and dataset name (where
    applicable), e.g. central-xnat//MYXNATPROJECT:pass_t1w_qc
name
    The name the source will be referenced by
datatype
    The data type of the column. Can be a field (int|float|str|bool),
    field array (list[int|float|str|bool]) or "file-group"
    (file, file+header/side-cars or directory)
""",
)
@click.argument("dataset_path")
@click.argument("name")
@click.argument("datatype")
@click.option(
    "--row-frequency",
    "-f",
    metavar="<dimension>",
    help=(
        "The row-frequency that items appear in the dataset (e.g. per "
        "'session', 'subject', 'timepoint', 'group', 'dataset' for "
        "medimage:Clinical data dimensions"
    ),
    show_default="highest",
)
@click.option(
    "--path",
    "-p",
    help=(
        "Path to item in the dataset. If 'regex' option is provided it will "
        "be treated as a regular-expression (in Python syntax)"
    ),
)
@click.option(
    "--salience",
    "-s",
    help=(
        "The salience of the column, i.e. whether it will show up on "
        "'arcana derive menu'"
    ),
)
def add_sink(dataset_path, name, datatype, row_frequency, path, salience):
    dataset = Dataset.load(dataset_path)
    dataset.add_sink(
        name=name,
        path=path,
        datatype=ClassResolver(DataType)(datatype),
        row_frequency=row_frequency,
        salience=salience,
    )
    dataset.save()


@dataset.command(
    name="missing-items",
    help="""Finds the IDs of rows that are missing a valid entry for an item in
the column.

Arguments
---------
dataset_path
    The path to the dataset including store and dataset name (where
    applicable), e.g. central-xnat//MYXNATPROJECT:pass_t1w_qc
name
    The name of the column to check
""",
)
@click.argument("dataset_path")
@click.argument("name")
def missing_items(name):
    raise NotImplementedError



from functools import reduce
from operator import mul
import pytest
from arcana.core.utils.testing import make_dataset_locator
from arcana.core.utils.testing.data import (
    make_dataset,
    TestDatasetBlueprint,
    TestDataSpace,
)
from arcana.dirtree.data import Text
from arcana.core.utils.testing.data import EncodedText
from arcana.core.deploy.command.base import ContainerCommand
from arcana.core.exceptions import ArcanaDataMatchError


def test_command_execute(concatenate_task, saved_dataset, work_dir):
    # Get CLI name for dataset (i.e. file system path prepended by 'file//')
    dataset_locator = make_dataset_locator(saved_dataset)
    bp = saved_dataset.__annotations__["blueprint"]
    duplicates = 1

    command_spec = ContainerCommand(
        task="arcana.core.utils.testing.tasks:" + concatenate_task.__name__,
        row_frequency=bp.space.default(),
        inputs=[
            {
                "name": "source1",
                "datatype": "common:Text",
                "field": "in_file1",
                "help_string": "dummy",
            },
            {
                "name": "source2",
                "datatype": "common:Text",
                "field": "in_file2",
                "help_string": "dummy",
            },
        ],
        outputs=[
            {
                "name": "sink1",
                "datatype": "common:Text",
                "field": "out_file",
                "help_string": "dummy",
            }
        ],
        parameters=[
            {
                "name": "duplicates",
                "datatype": "int",
                "default": 2,
                "help_string": "dummy",
            }
        ],
    )
    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    command_spec.execute(
        dataset_locator=dataset_locator,
        input_values=[
            ("source1", "file1"),
            ("source2", "file2"),
        ],
        output_values=[
            ("sink1", "concatenated"),
        ],
        parameter_values=[
            ("duplicates", str(duplicates)),
        ],
        raise_errors=True,
        plugin="serial",
        work_dir=str(work_dir),
        loglevel="debug",
        dataset_hierarchy=",".join(str(ln) for ln in bp.hierarchy),
        pipeline_name="test_pipeline",
    )
    # Add source column to saved dataset
    sink = saved_dataset.add_sink("concatenated", Text)
    assert len(sink) == reduce(mul, bp.dim_lengths)
    fnames = ["file1.txt", "file2.txt"]
    if concatenate_task.__name__.endswith("reverse"):
        fnames = [f[::-1] for f in fnames]
    expected_contents = "\n".join(fnames * duplicates)
    for item in sink:
        item.get(assume_exists=True)
        with open(item.fs_path) as f:
            contents = f.read()
        assert contents == expected_contents


def test_command_execute_fail(concatenate_task, saved_dataset, work_dir):
    # Get CLI name for dataset (i.e. file system path prepended by 'file//')
    dataset_locator = make_dataset_locator(saved_dataset)
    bp = saved_dataset.__annotations__["blueprint"]
    duplicates = 1

    command_spec = ContainerCommand(
        task="arcana.core.utils.testing.tasks:" + concatenate_task.__name__,
        row_frequency=bp.space.default(),
        inputs=[
            {
                "name": "source1",
                "datatype": "common:Text",
                "field": "in_file1",
                "help_string": "dummy",
            },
            {
                "name": "source2",
                "datatype": "common:Directory",
                "field": "in_file2",
                "help_string": "dummy",
            },
        ],
        outputs=[
            {
                "name": "sink1",
                "datatype": "common:Text",
                "field": "out_file",
                "help_string": "dummy",
            }
        ],
        parameters=[
            {
                "name": "duplicates",
                "datatype": "int",
                "default": 2,
                "help_string": "dummy",
            }
        ],
    )

    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    with pytest.raises(ArcanaDataMatchError):
        command_spec.execute(
            dataset_locator=dataset_locator,
            input_values=[
                ("source1", "bad-file-path"),
                ("source2", "file2"),
            ],
            output_values=[
                ("sink1", "concatenated"),
            ],
            parameter_values=[
                ("duplicates", duplicates),
            ],
            raise_errors=True,
            plugin="serial",
            work_dir=str(work_dir),
            loglevel="debug",
            dataset_hierarchy=",".join(str(ln) for ln in bp.hierarchy),
            pipeline_name="test_pipeline",
        )


def test_command_execute_on_row(cli_runner, work_dir):

    # Create test dataset consisting of a single row with a range of filenames
    # from 0 to 4
    filenumbers = list(range(5))
    bp = TestDatasetBlueprint(
        [
            TestDataSpace.abcd
        ],  # e.g. XNAT where session ID is unique in project but final layer is organised by timepoint
        [1, 1, 1, 1],
        [f"{i}.txt" for i in filenumbers],
        {},
        {},
        [],
    )
    dataset_path = work_dir / "numbered_dataset"
    dataset = make_dataset(bp, dataset_path)
    dataset.save()

    # Get CLI name for dataset (i.e. file system path prepended by 'file//')
    dataset_locator = make_dataset_locator(dataset)

    def get_dataset_filenumbers():
        dataset.refresh()
        row = next(dataset.rows())
        return sorted(int(i.path) for i in row.unresolved)

    assert get_dataset_filenumbers() == filenumbers

    command_spec = ContainerCommand(
        task="arcana.core.utils.testing.tasks:plus_10_to_filenumbers",
        row_frequency=bp.space.default(),
        inputs=[
            {
                "name": "a_row",
                "datatype": "arcana.core.data.row:DataRow",
                "field": "filenumber_row",
                "help_string": "dummy",
            },
        ],
    )

    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    command_spec.execute(
        dataset_locator=dataset_locator,
        input_values=[
            ("a_row", ""),
        ],
        raise_errors=True,
        plugin="serial",
        work_dir=str(work_dir),
        loglevel="debug",
        dataset_hierarchy=",".join(str(ln) for ln in bp.hierarchy),
        pipeline_name="test_pipeline",
    )

    assert get_dataset_filenumbers() == [i + 10 for i in filenumbers]


def test_command_execute_with_converter_args(saved_dataset, work_dir):
    """Test passing arguments to file format converter tasks via input/output
    "qualifiers", e.g. 'converter.shift=3' using the arcana-run-pipeline CLI
    tool (as used in the XNAT CS commands)
    """
    # Get CLI name for dataset (i.e. file system path prepended by 'file//')
    dataset_locator = make_dataset_locator(saved_dataset)
    bp = saved_dataset.__annotations__["blueprint"]
    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    command_spec = ContainerCommand(
        task="arcana.core.utils.testing.tasks:identity_file",
        row_frequency=bp.space.default(),
        inputs=[
            {
                "name": "source",
                "datatype": "arcana.core.utils.testing.data:EncodedText",
                "default_column": {"datatype": "common:Text"},
                "field": "in_file",
                "help_string": "dummy",
            },
        ],
        outputs=[
            {
                "name": "sink1",
                "datatype": "arcana.core.utils.testing.data:EncodedText",
                "field": "out",
                "help_string": "dummy",
            },
            {
                "name": "sink2",
                "datatype": "arcana.core.utils.testing.data:EncodedText",
                "default_column": {
                    "datatype": "arcana.core.utils.testing.data:DecodedText"
                },
                "field": "out",
                "help_string": "dummy",
            },
        ],
    )

    command_spec.execute(
        dataset_locator=dataset_locator,
        input_values=[
            ("source", "file1 converter.shift=3"),
        ],
        output_values=[
            ("sink1", "encoded"),
            ("sink2", "decoded converter.shift=3"),
        ],
        raise_errors=True,
        plugin="serial",
        work_dir=str(work_dir),
        loglevel="debug",
        dataset_hierarchy=",".join(str(ln) for ln in bp.hierarchy),
        pipeline_name="test_pipeline",
    )
    # Add source column to saved dataset
    saved_dataset.add_sink("sink1", EncodedText, path="encoded")
    saved_dataset.add_sink("sink2", Text, path="decoded")
    unencoded_contents = "file1.txt"
    encoded_contents = (
        "iloh41w{w"  # 'file1.txt' characters shifted up by 3 in ASCII code
    )
    for row in saved_dataset.rows(frequency="abcd"):
        enc_item = row["sink1"]
        dec_item = row["sink2"]
        enc_item.get(assume_exists=True)
        dec_item.get(assume_exists=True)
        with open(enc_item.fs_path) as f:
            enc_contents = f.read()
        with open(dec_item.fs_path) as f:
            dec_contents = f.read()
        assert enc_contents == encoded_contents
        assert dec_contents == unencoded_contents



import os
from unittest.mock import patch
from arcana.core.cli.store import add, ls, remove, rename
from arcana.core.utils.testing import show_cli_trace
from arcana.core.data.store import DataStore

STORE_URI = "http://dummy.uri"
STORE_USER = "a_user"
STORE_PASSWORD = "a-password"


def test_store_cli(cli_runner, work_dir):
    test_home_dir = work_dir / "test-arcana-home"
    store_name = "test-mock"
    # Create a new home directory so it doesn't conflict with user settings
    with patch.dict(os.environ, {"ARCANA_HOME": str(test_home_dir)}):
        # Add new XNAT configuration
        result = cli_runner(
            add,
            [
                store_name,
                "arcana.core.utils.testing.data:MockDataStore",
                STORE_URI,
                "--user",
                STORE_USER,
                "--password",
                STORE_PASSWORD,
            ],
        )
        assert result.exit_code == 0, show_cli_trace(result)
        # List all saved and built-in stores
        result = cli_runner(ls, [])
        assert result.exit_code == 0, show_cli_trace(result)
        assert "file - arcana.dirtree.data.file_system:FileSystem" in result.output
        assert (
            f"{store_name} - arcana.core.utils.testing.data:MockDataStore"
            in result.output
        )
        assert "    server: " + STORE_URI in result.output


def test_store_cli_remove(cli_runner, work_dir):
    test_home_dir = work_dir / "test-arcana-home"
    new_store_name = "a-new-mock"
    # Create a new home directory so it doesn't conflict with user settings
    with patch.dict(os.environ, {"ARCANA_HOME": str(test_home_dir)}):
        # Add new XNAT configuration
        cli_runner(
            add,
            [
                new_store_name,
                "arcana.core.utils.testing.data:MockDataStore",
                STORE_URI,
                "--user",
                STORE_USER,
                "--password",
                STORE_PASSWORD,
            ],
        )
        # Check store is saved
        result = cli_runner(ls, [])
        assert new_store_name in result.output

        cli_runner(remove, new_store_name)
        # Check store is gone
        result = cli_runner(ls, [])
        assert new_store_name not in result.output


def test_store_cli_rename(cli_runner, work_dir):
    test_home_dir = work_dir / "test-arcana-home"
    old_store_name = "i123"
    new_store_name = "y456"
    # Create a new home directory so it doesn't conflict with user settings
    with patch.dict(os.environ, {"ARCANA_HOME": str(test_home_dir)}):
        # Add new XNAT configuration
        cli_runner(
            add,
            [
                old_store_name,
                "arcana.core.utils.testing.data:MockDataStore",
                STORE_URI,
                "--user",
                STORE_USER,
                "--password",
                STORE_PASSWORD,
            ],
        )
        # Check store is saved
        result = cli_runner(ls, [])
        assert "i123 - arcana.core.utils.testing.data:MockDataStore" in result.output

        cli_runner(rename, [old_store_name, new_store_name])
        # Check store is renamed
        result = cli_runner(ls, [])
        assert (
            "i123 - arcana.core.utils.testing.data:MockDataStore" not in result.output
        )
        assert "y456 - arcana.core.utils.testing.data:MockDataStore" in result.output


def test_store_cli_encrypt_credentials(cli_runner, work_dir):
    test_home_dir = work_dir / "test-arcana-home"
    store_name = "another-test-mock"
    # Create a new home directory so it doesn't conflict with user settings
    with patch.dict(os.environ, {"ARCANA_HOME": str(test_home_dir)}):
        # Add new XNAT configuration
        result = cli_runner(
            add,
            [
                store_name,
                "arcana.core.utils.testing.data:MockDataStore",
                STORE_URI,
                "--user",
                STORE_USER,
                "--password",
                STORE_PASSWORD,
            ],
        )
        assert result.exit_code == 0, show_cli_trace(result)
        # Check credentials have been encrypted
        loaded_store = DataStore.load(store_name)
        assert loaded_store.password != ""
        assert loaded_store.password is not STORE_PASSWORD
        assert loaded_store.user != ""
        assert loaded_store.user is not STORE_USER
